'''
Default settings
'''

__all__ = 'ITEMS',

ITEMS = (
    ('bow', (0, 0), 2,
     {'Bow and Arrows': 'bow', 'Bow and Silver Arrows': 'bowandsilvers'},
     {'links': {'silverarrows': (2,)}}),
    ('silverarrows', (), 1, {}, {}),
    ('boomerang', (0, 1), 3,
     {'Boomerang': 'blueboomerang', 'Magical Boomerang': 'redboomerang',
      'Boomerangs': 'bothboomerangs'},
     {'links': {'blueboomerang': (1, 3), 'redboomerang': (2, 3)}}),
    ('blueboomerang', (), 1, {}, {}),
    ('redboomerang', (), 1, {}, {}),
    ('hookshot', (0, 2), 1, {'Hookshot': 'hookshot'}, {}),
    ('bombs', (0, 3), 1, {'Bombs': 'bombs'}, {'default': 1}),
    ('mushroompowder', (0, 4), 3,
     {'Mushroom': 'mushroom', 'Magic Powder': 'powder',
      'Mushroom/Magic Powder': 'mushroompowder'},
     {'links': {'mushroom': (1, 3), 'powder': (2, 3)}}),
    ('mushroom', (), 1, {}, {}),
    ('powder', (), 1, {}, {}),
    ('firerod', (1, 0), 1, {'Fire Rod': 'firerod'}, {}),
    ('icerod', (1, 1), 1, {'Ice Rod': 'icerod'}, {}),
    ('bombos', (1, 2), 1, {'Bombos Medallion': 'bombos'}, {}),
    ('ether', (1, 3), 1, {'Ether Medallion': 'ether'}, {}),
    ('quake', (1, 4), 1, {'Quake Medallion': 'quake'}, {}),
    ('lantern', (2, 0), 1, {'Lantern': 'lantern'}, {}),
    ('hammer', (2, 1), 1, {'Magic Hammer': 'hammer'}, {}),
    ('shovelocarina', (2, 2), 3,
     {'Shovel': 'shovel', 'Ocarina': 'ocarina',
      'Shovel/Ocarina': 'shovelocarina'},
     {'links': {'shovel': (1, 3), 'ocarina': (2, 3)}}),
    ('shovel', (), 1, {}, {}),
    ('ocarina', (), 1, {}, {}),
    ('bugnet', (2, 3), 1, {'Bug-Catching Net': 'bugnet'}, {}),
    ('mudora', (2, 4), 1, {'Book of Mudora': 'mudora'}, {}),
    ('bottle', (3, 0), 1, {'Bottle': 'bottle'}, {}),
    ('somaria', (3, 1), 1, {'Cane of Somaria': 'somaria'}, {}),
    ('byrna', (3, 2), 1, {'Cane of Byrna': 'byrna'}, {}),
    ('cape', (3, 3), 1, {'Magic Cape': 'cape'}, {}),
    ('mirror', (3, 4), 1, {'Magic Mirror': 'mirror'}, {}),

    ('sword', (1, 5), 4,
     {"Fighter's Sword": 'sword0', 'Master Sword': 'sword1',
      'Tempered Sword': 'sword2', 'Golden Sword': 'sword3'},
     {'links': {'mastersword': (2, 3, 4), 'mastersword2': (3, 4)}}),
    ('mastersword', (), 1, {}, {}),
    ('mastersword2', (), 1, {}, {}),
    ('shield', (2, 5), 3,
     {"Fighter's Shield": 'shield1', 'Red Shield': 'shield2',
      'Mirror Shield': 'shield3'},
     {'links': {'mirrorshield': (3,)}}),
    ('mirrorshield', (), 1, {}, {}),
    ('armour', (3, 5), 3,
     {'Green Tunic': 'armour1', 'Blue Mail': 'armour2',
      'Red Mail': 'armour3'},
     {'default': 1, 'links': {'bluemail': (2, 3), 'redmail': (3,)}}),
    ('bluemail', (), 1, {}, {}),
    ('redmail', (), 1, {}, {}),

    ('pegasus', (4, 1), 1, {'Pegasus Shoes': 'pegasus'}, {}),
    ('glove', (4, 2), 2,
     {'Power Gloves': 'glove1', "Titan's Mitts": 'glove2'},
     {'links': {'powerglove': (1, 2), 'titansmitts': (2,)}}),
    ('powerglove', (), 1, {}, {}),
    ('titansmitts', (), 1, {}, {}),
    ('flippers', (4, 3), 1, {"Zora's Flippers": 'flippers'}, {}),
    ('pearl', (4, 4), 1, {'Moon Pearl': 'pearl'}, {}),
    ('halfmagic', (4, 5), 1, {'Half-Magic': 'halfmagic'}, {})
)



'''
Location map for Zelda 3
'''

from .config import CONFIG
from . import gui

__all__ = 'main',


def main() -> None:
    '''
    Main program.
    '''

    gui.GUI = gui.guilib.GraphicalInterface()
    gui.GUI.run()



'''
Image files
'''

import os.path
import tkinter as tk

__all__ = 'image', 'makebw'


def image(filename: str) -> str:
    '''
    Return full path to image file.

    Args:
        file: basename of image file
    Returns:
        str: full path to image file
    '''

    imagedir = os.path.join(os.path.dirname(__file__), '../images')
    imagepath = os.path.join(imagedir, '{0:s}.png'.format(filename))
    return os.path.normpath(imagepath)


def makebw(img: tk.PhotoImage) -> tk.PhotoImage:
    '''
    Make black&white version of image

    Args:
        img: image to convert
    '''

    for x in range(img.width()):
        for y in range(img.height()):
            bw = sum(img.get(x, y)) // 3
            if bw in (0, 255):
                continue
            img.put('#{0:02x}{0:02x}{0:02x}'.format(bw), (x, y))
    return img



'''
Dungeon display.
'''

import importlib
import tkinter as tk
import tkinter.ttk as ttk
import typing

from ..config import CONFIG
from ..config.images import image, makebw
from .. import dungeons
common = importlib.import_module(
    '..gui-common.dungeons', package=__package__)
REWARDS = common.REWARDS

from . import misc

__all__ = 'DungeonWindow',


class DungeonWindow(tk.Toplevel):
    '''
    Dungeon check display.

    Instance variables:
        tracker: dungeon tracker object
        layout: dungeon layout in display
        helpertext: helper text variable
        scaling: up- and downscale factors for objects
        widgets: single dungeon button collections
    '''

    def __init__(self, tracker):
        '''
        Args:
            tracker: dungeon tracker object
        '''

        super().__init__()
        self.title('Dungeons')

        self.tracker = tracker

        self.frame = ttk.Frame(self)
        self.frame.pack(side='top', fill='both', expand=True)

        self.bottomframe = ttk.Frame(self)
        self.bottomframe.pack(side='top', fill='x', expand=True)
        self.helperframe = ttk.Frame(self.bottomframe)
        self.helperframe.pack(side='left', fill='x', expand=True)
        self.helpertext = tk.StringVar()
        self.helper = ttk.Label(
            self.helperframe, anchor=tk.CENTER, textvariable=self.helpertext)
        self.helper.pack(side='left', fill='x', expand=True)

        self.scaling = _scale_factors()
        scale = self.scaling[0] / self.scaling[1]
        _icon_scale = lambda icon: icon.zoom(
            self.scaling[0], self.scaling[0]).subsample(
                self.scaling[1], self.scaling[1])

        self.ganoncrystals = {
            'icon': _icon_scale(
                tk.PhotoImage(
                    file=image('ganonstower'),
                    master=self.bottomframe)).subsample(2, 2),
            'canvas': tk.Canvas(
                self.bottomframe, height=32 * scale, width=48 * scale),
            'count': -1}
        self.ganoncrystals['image'] = self.ganoncrystals['canvas'].create_image(
            0, 0, anchor=tk.NW, image=self.ganoncrystals['icon'])
        self.ganoncrystals['text'] = self.ganoncrystals['canvas'].create_text(
            48 * scale, 32 * scale, anchor=tk.SE, text='?')
        self.ganoncrystals['canvas'].bind(
            '<Enter>', lambda _: self.helpertext.set("Ganon's Pyramid Access"))
        self.ganoncrystals['canvas'].bind(
            '<Leave>', lambda _: self.helpertext.set(''))
        self.ganoncrystals['canvas'].bind(
            '<ButtonRelease-1>', lambda _: self.set_crystal('ganon', +1))
        self.ganoncrystals['canvas'].bind(
            '<ButtonRelease-3>', lambda _: self.set_crystal('ganon', -1))
        self.ganoncrystals['canvas'].pack(side='right')
        self.set_crystal('ganon', 0)

        self.towercrystals = {
            'icon': _icon_scale(
                tk.PhotoImage(
                    file=image('reward-crystal'), master=self.bottomframe)),
            'canvas': tk.Canvas(
                self.bottomframe, height=32 * scale, width=48 * scale),
            'count': -1}
        self.towercrystals['image'] = self.towercrystals['canvas'].create_image(
            0, 0, anchor=tk.NW, image=self.towercrystals['icon'])
        self.towercrystals['text'] = self.towercrystals['canvas'].create_text(
            48 * scale, 32 * scale, anchor=tk.SE, text='?')
        self.towercrystals['canvas'].bind(
            '<Enter>', lambda _: self.helpertext.set("Ganon's Tower Access"))
        self.towercrystals['canvas'].bind(
            '<Leave>', lambda _: self.helpertext.set(''))
        self.towercrystals['canvas'].bind(
            '<ButtonRelease-1>', lambda _: self.set_crystal('tower', +1))
        self.towercrystals['canvas'].bind(
            '<ButtonRelease-3>', lambda _: self.set_crystal('tower', -1))
        self.towercrystals['canvas'].pack(side='right')
        self.set_crystal('tower', 0)

        buttonstyle = ttk.Style()
        buttonstyle.configure('Dungeonbutton.TButton', relief=tk.FLAT)

        self.widgets = []
        self.redraw()

    def _dungeon_display(self, dungeon) -> None:
        '''
        Make and place single dungeons display object.

        Args:
            dungeon: dungeon object
        Writes:
            buttons
        '''

        widget = Dungeon(dungeon, self.frame, self.scaling)
        widget.bind(
            '<Enter>', lambda _: self.helpertext.set(dungeon.identifier))
        widget.bind(
            '<Leave>', lambda _: self.helpertext.set(''))

        for button in widget.buttons:
            button.bind('<ButtonRelease-1>', self.tracker.save, add='+')
            button.bind('<ButtonRelease-3>', self.tracker.save, add='+')
            button.bind(
                '<ButtonRelease-1>', lambda _: widget.check_state(dungeon),
                add='+')
            button.bind(
                '<ButtonRelease-3>', lambda _: widget.check_state(dungeon),
                add='+')

        widget.grid(
            column=dungeon.location[0], row=dungeon.location[1], sticky=misc.A)
        self.widgets.append(widget)

    def reset(self) -> None:
        '''
        Reset dungeons to default.
        '''

        for cobj in ('tower', 'ganon'):
            self.set_crystal(cobj, None)
        self.tracker.reset()
        self.redraw()

    def redraw(self) -> None:
        '''
        Recreate dungeon display.
        '''

        while self.widgets:
            self.widgets.pop().destroy()
        for dungeon in self.tracker:
            self._dungeon_display(self.tracker[dungeon])

    def set_crystal(self, ctype: str, diff: int) -> None:
        '''
        Set crystal requirements.

        Args:
            ctype: 'tower' or 'crystal'
            diff: -1, 0, +1 or None (for reset)
        '''

        assert ctype in ('tower', 'ganon')
        assert diff in (-1, 0, +1, None)
        scale = self.scaling[0] / self.scaling[1]
        cobj = self.towercrystals if ctype == 'tower' else self.ganoncrystals
        cobj['count'] = self.tracker.set_crystal(ctype, diff)
        if cobj['count'] == -1:
            newtext = '?'
        else:
            newtext = str(cobj['count'])
        cobj['canvas'].delete(cobj['text'])
        cobj['text'] = cobj['canvas'].create_text(
            48 * scale, 32 * scale, anchor=tk.SE, font=('Arial Black', 16),
            text=newtext)


class Dungeon(ttk.Frame):
    '''
    Single dungeon display object.
    '''

    def __init__(
            self, dungeon, parent: ttk.Frame, scaling: (int, int)):
        '''
        dungeon: dungeon object
        parent: parent widget for object
        scaling: button scale
        '''

        self.scaling = scaling
        scale = scaling[0] / scaling[1]

        super().__init__(parent, borderwidth=2, relief=tk.RIDGE)
        self.child = ttk.Label(self)
        self.child.grid(column=0, row=0, sticky=misc.A)

        icon = self._icon_scale(
            tk.PhotoImage(file=dungeon.icon, master=parent)
            if dungeon.icon else None)
        self.pic = ttk.Label(
            self.child, borderwidth=1, image=icon, relief=tk.RAISED)
        self.pic.grid(column=0, row=0, columnspan=2, rowspan=2)
        self.icon = icon

        self.rewardname = dungeon.get('reward')
        self.rewardimg = REWARDS[self.rewardname]
        self.rewardicon = self._icon_scale(
            tk.PhotoImage(file=image(self.rewardimg), master=self))
        self.reward = tk.Canvas(
            self.child, height=32 * scale, width=32 * scale)
        if 'reward' in dungeon.features:
            self.rewardid = self.reward.create_image(
                0, 0, anchor=tk.NW, image=self.rewardicon)
            self.reward.bind(
                '<ButtonRelease-1>', lambda _: dungeon.cycle_reward(True))
            self.reward.bind(
                '<ButtonRelease-3>', lambda _: dungeon.cycle_reward(False))
        self.reward.grid(column=2, row=0)

        self.complete = tk.Canvas(
            self.child, height=32 * scale, width=32 * scale)
        self.complete.create_rectangle(5, 5, 27, 27, width=4)
        self.completeid = None
        self.complete.bind(
            '<ButtonRelease-1>', lambda _: dungeon.mark_complete(True))
        self.complete.bind(
            '<ButtonRelease-3>', lambda _: dungeon.mark_complete(False))
        self.complete.grid(column=2, row=1)

        self.medallionname = dungeon.medallion
        self.medallionicon = self._icon_scale(
            tk.PhotoImage(
                file=image('medallion-{0:s}'.format(self.medallionname)),
                master=self))
        self.medallion = tk.Canvas(
            self.child, height=32 * scale, width=32 * scale)
        if 'medallion' in dungeon.features:
            self.medallionid = self.medallion.create_image(
                0, 0, anchor=tk.NW, image=self.medallionicon)
            self.medallion.bind(
                '<ButtonRelease-1>', lambda _: dungeon.cycle_medallion(True))
            self.medallion.bind(
                '<ButtonRelease-3>', lambda _: dungeon.cycle_medallion(False))
        self.medallion.grid(column=3, row=0)

        self.bigkeyicon = self._icon_scale(
            tk.PhotoImage(file=image('bigkey'), master=parent))
        self.bigkey = tk.Canvas(
            self.child, height=32 * scale, width=32 * scale)
        if ('bigkey' in dungeon.features and
            CONFIG['dungeon_items'] == 'Keysanity'):
            self.bigkeyid = self.bigkey.create_image(
                0, 0, anchor=tk.NW, image=self.bigkeyicon)
            self.bigkey.bind('<ButtonRelease-1>', dungeon.toggle_bigkey)
            self.bigkey.bind('<ButtonRelease-3>', dungeon.toggle_bigkey)
        self.bigkey.grid(column=3, row=1, sticky=tk.E)

        self.keyicon = self._icon_scale(
            tk.PhotoImage(file=image('smallkey'), master=parent))
        self.key = tk.Canvas(
            self.child, height=32 * scale, width=48 * scale)
        if (dungeon.totalkeys > 0 and
            CONFIG['dungeon_items'] in (
                'Maps/Compasses/Small Keys', 'Keysanity') and
            CONFIG['world_state'] != 'Retro'):
            self.keyimg = self.key.create_image(
                0, 0, anchor=tk.NW, image=self.keyicon)
            self.keytext = self.key.create_text(
                48 * scale, 32 * scale, anchor=tk.SE, text='0')
            self.key.bind('<ButtonRelease-1>', dungeon.key_up)
            self.key.bind('<ButtonRelease-3>', dungeon.key_down)
        self.key.grid(column=4, columnspan=2, row=0)

        self.itemicon = self._icon_scale(
            tk.PhotoImage(file=image('chest_full'), master=parent))
        self.item = tk.Canvas(
            self.child, height=32 * scale, width=48 * scale)
        if dungeon.total_items() > 0:
            self.itemimg = self.item.create_image(
                0, 0, anchor=tk.NW, image=self.itemicon)
            self.itemtext = self.item.create_text(
                48 * scale, 32 * scale, anchor=tk.SE, text='0')
            self.item.bind('<ButtonRelease-1>', dungeon.item_up)
            self.item.bind('<ButtonRelease-3>', dungeon.item_down)
        if CONFIG['dungeon_items'] in (
                'Maps/Compasses/Small Keys', 'Keysanity'):
            self.item.grid(column=4, columnspan=2, row=1)
        else:
            self.item.grid(column=3, columnspan=2, row=1)

        self.buttons = (
            self.reward, self.complete, self.medallion, self.bigkey, self.key,
            self.item)

        self.check_state(dungeon)

    def check_state(self, dungeon) -> None:
        '''
        Check button state and make adjustments if necessary.

        Args:
            dungeon: dungeon object
        '''

        # Check whether reward image needs to be changed.
        if self.rewardname != dungeon.reward:
            self.rewardname = dungeon.reward
            self.rewardimg = REWARDS[dungeon.reward]
            self.rewardicon = self._icon_scale(
                tk.PhotoImage(file=image(self.rewardimg), master=self))
            self.reward.delete(self.rewardid)
            self.rewardid = self.reward.create_image(
                0, 0, anchor=tk.NW, image=self.rewardicon)

        # Check completion.
        icon = self._icon_scale(tk.PhotoImage(file=dungeon.icon, master=self))
        if dungeon.completed:
            icon = makebw(icon)
        if dungeon.completed and self.completeid is None:
            self.completeid = self.complete.create_text(
                17, 17, anchor=tk.CENTER, font=('Arial Black', 20), text='✔')
        elif not dungeon.completed and self.completeid is not None:
            self.complete.delete(self.completeid)
            self.completeid = None
        self.pic = ttk.Label(
            self.child, borderwidth=1, image=icon, relief=tk.RAISED)
        self.pic.grid(column=0, row=0, columnspan=2, rowspan=2)
        self.icon = icon

        # Check whether medallion image needs to be changed.
        if self.medallionname != dungeon.medallion:
            self.medallionname = dungeon.medallion
            self.medallionicon = self._icon_scale(
                tk.PhotoImage(
                    file=image('medallion-{0:s}'.format(self.medallionname)),
                    master=self))
            self.medallion.delete(self.medallionid)
            self.medallionid - self.medallion.create_image(
                0, 0, anchor=tk.NW, image=self.medallionicon)
            

        # Check whether the bigkey button should be disabled.
        try:
            self.bigkey.delete(self.bigkeyid)
        except AttributeError:
            pass
        else:
            self.bigkeyicon = self._icon_scale(
                tk.PhotoImage(file=image('bigkey'), master=self))
            if not dungeon.bigkey:
                self.bigkeyicon = makebw(self.bigkeyicon)
            self.bigkey.create_image(
                0, 0, anchor=tk.NW, image=self.bigkeyicon)

        # Check numbers (small keys).
        scale = self.scaling[0] / self.scaling[1]
        try:
            self.key.delete(self.keytext)
        except AttributeError:
            pass
        else:
            self.keytext = self.key.create_text(
                48 * scale, 32 * scale, anchor=tk.SE,
                font=('Arial Black', int(16 * scale)),
                text=str(dungeon.smallkeys))

        # Check numbers (items).
        try:
            self.item.delete(self.itemtext)
        except AttributeError:
            pass
        else:
            itemfont = (
                ('Arial Black', int(8 * scale)) if dungeon.remaining() > 9
                else ('Arial Black', int(16 * scale)))
            self.itemtext = self.item.create_text(
                48 * scale, 32 * scale, anchor=tk.SE, font=itemfont,
                text=str(dungeon.remaining()))
            newchest = (
                'chest_full' if dungeon.remaining() > 0 else 'chest_empty')
            self.itemicon = self._icon_scale(
                tk.PhotoImage(file=image(newchest), master=self))
            self.item.delete(self.itemimg)
            self.itemimg = self.item.create_image(
                0, 0, anchor=tk.NW, image=self.itemicon)

    def _icon_scale(self, icon: tk.PhotoImage) -> tk.PhotoImage:
        '''
        Rescale icon.

        Args:
            icon: icon to be rescaled
        Returns:
            tk.PhotoImage: rescaled icon
        '''

        if self.scaling[0] != 1:
            icon = icon.zoom(self.scaling[0], self.scaling[0])
        if self.scaling[1] != 1:
            icon = icon.subsample(self.scaling[1], self.scaling[1])
        return icon


def _scale_factors() -> (int, int):
    '''
    Calculate up- and downscale factor.

    Returns:
        int: upscale factor
        int: downscale factor
    '''

    scaling = CONFIG['icon_size']
    for up in range(1, 1000):
        if not (scaling * up) % 1:
            upscale = int(scaling * up)
            break
    else:
        CONFIG.set('icon_size', 1)
        assert False
    downscale = int(upscale // scaling)
    return upscale, downscale


def start(worldtracker) -> DungeonWindow:
    '''
    Initialise dungeon tracker window.

    Args:
        worldtracker: world state tracker
    '''

    return DungeonWindow(dungeons.DungeonTracker(worldtracker))



'''
Light World map
'''

from . import maps

__all__ = 'start',


def start(worldtracker) -> maps.MapDisplay:
    '''
    Open light world map.

    Args:
        worldtracker: world state tracker
    '''

    display = maps.MapDisplay('light')
    worldtracker.add_map('light', display)
    return display



'''
Eastern Dark World
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Dark River': {
        'type': 'area',
        'link': {
            'Little River': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Witch Hut River': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'River Shop Area': [],
            'Dark World Waterway': []}
    },
    'River Shop Area': {
        'type': 'area',
        'link': {
            'East Light World': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'East Dark World': [('item', 'hammer'), ('item', 'powerglove')],
            'Dark River': [('item', 'flippers')],
            'North Dark World': [('item', 'hookshot')],
            'River Shop Entrance (E)': [],
            'Catfish Approach': [('item', 'powerglove')]}
    },
    'River Shop Entrance (E)': {
        'type': 'entrance_shop', 'map': 'dark', 'coord': (528, 224),
        'link': {
            'River Shop Area': [],
            'River Shop Entrance (I)': []}
    },
    'River Shop Entrance (I)': {
        'type': 'interior',
        'link': {
            'River Shop Entrance (E)': []}
    },
    'Catfish Approach': {
        'type': 'area',
        'link': {
            "Zora's River": [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'River Shop Area': [('item', 'powerglove')],
            'Dark River': [('item', 'flippers')],
            'Catfish': [('rabbitbarrier', None)]}
    },
    'Catfish': {
        'type': 'item', 'map': 'dark', 'coord': (587, 114),
        'link': {
            'Catfish Approach': []}
    },

    'East Dark World': {
        'type': 'area',
        'link': {
            'East Light World': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Dark Bottle Ocarina': [],
            'River Shop Area': [
                ('item', 'hammer'), ('item', 'powerglove'),
                ('and', [
                    ('glitch', 'overworld'), ('item', 'pegasus')])],
            'Dark River': [('item', 'flippers')],
            'Pyramid': [],
            'South Dark World': [('item', 'hammer')],
            'Frozen Lake': [('item', 'flippers')],
            'Dark Fairy Cave Entrance (E)': [],
            'Fake Dark Palace Entrance (E)': [],
            'East Portal (DW)': [('item', 'hammer')],
            'Dark Fairy Pond Entrance (E)': [],
            'Dark Palace Entrance (E)': [('rabbitbarrier', None)]}
    },
    'Dark Fairy Cave Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (540, 427),
        'link': {
            'East Dark World': [],
            'Dark Fairy Cave Entrance (I)': []}
    },
    'Dark Fairy Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Dark Fairy Cave Entrance (E)': []}
    },
    'Fake Dark Palace Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (557, 332),
        'link': {
            'East Dark World': [],
            'Fake Dark Palace Entrance (I)': []}
    },
    'Fake Dark Palace Entrance (I)': {
        'type': 'interior',
        'link': {
            'Fake Dark Palace Entrance (E)': []}
    },
    'East Portal (DW)': {
        'type': 'area',
        'link': {
            'East Portal (LW)': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'East Portal': [('item', 'powerglove')],
            'East Dark World': [('item', 'hammer')]}
    },
    'Dark Fairy Pond Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (644, 464),
        'link': {
            'East Dark World': [],
            'Dark Fairy Pond Entrance (I)': []}
    },
    'Dark Fairy Pond Entrance (I)': {
        'type': 'interior',
        'link': {
            'Dark Fairy Pond Entrance (I)': []}
    },

    'Pyramid': {
        'type': 'area',
        'link': {
            'Hyrule Castle': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Castle Walls': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Dark Bottle Ocarina': [],
            'East Dark World': [],
            'Pyramid Item': [],
            'Pyramid Fairy Entrance (E)': [('and', [
                ('access', 'Bomb Shop Entrance (I)'), ('crystals', 'fairy'),
                ('rabbitbarrier', None)])],
            'Ganon Drop Entrance (E)': [('and', [
                ('nosettings', 'inverted'), ('macro', 'ganon')])]}
    },
    'Pyramid Item': {
        'type': 'item', 'map': 'dark', 'coord': (379, 297),
        'link': {
            'Pyramid': []}
    },
    'Pyramid Fairy Entrance (E)': {
        'type': 'entrance_unique', 'map': 'dark', 'coord': (305, 321),
        'link': {
            'Pyramid': [],
            'Pyramid Fairy Entrance (I)': []}
    },
    'Pyramid Fairy Entrance (I)': {
        'type': 'interior',
        'link': {
            'Pyramid Fairy Entrance (E)': [],
            'Pyramid Fairy': []}
    },
    'Pyramid Fairy': {
        'type': 'cave', 'map': 'dark', 'coord': (305, 321),
        'link': {
            'Pyramid Fairy Entrance (I)': []}
    },
    'Ganon Drop Entrance (E)': {
        'type': 'entrance_drop', 'map': 'dark', 'coord': (324, 270),
        'link': {
            'Ganon Drop Entrance (I)': []}
    },
    'Ganon Drop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Ganon': [('and', [
                ('item', 'mastersword'), ('item', 'bow'),
                ('item', 'silverarrows')])]}
    },
    'Ganon': {
        'type': 'ganon', 'map': 'dark', 'coord': (324, 240),
        'link': {}
    },
}



'''
Mountain Tower
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Mountain Tower Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'light', 'coord': (372, 35),
        'link': {
            'West Upper Mount Hebra': [],
            'Mountain Tower Entrance (I)': [('nosettings', 'inverted')],
            'Castle Tower Entrance (E)': [('settings', 'inverted')]}
    },
    'Mountain Tower Entrance (I)': {
        'type': 'interior',
        'link': {
            'Mountain Tower Entrance (E)': [],
            'Mountain Tower Interior': [('rabbitbarrier', None)]}
    },
    'Mountain Tower Interior': {
        'type': 'area', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Small Key': [],
            'Mountain Tower Map': [],
            'Mountain Tower Big Key Descent': [
                ('smallkey', ('Mountain Tower', 2))],
            'Mountain Tower Ascent': [('bigkey', 'Mountain Tower')]}
    },
    'Mountain Tower Small Key': {
        'type': 'dungeonchest', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Interior': []}
    },
    'Mountain Tower Map': {
        'type': 'dungeonchest', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Interior': []}
    },
    'Mountain Tower Big Key Descent': {
        'type': 'area', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Interior': [],
            'Mountain Tower Big Key': [
                ('item', 'lantern'), ('item', 'firerod')]}
    },
    'Mountain Tower Big Key': {
        'type': 'dungeonchest', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Big Key Descent': []}
    },
    'Mountain Tower Ascent': {
        'type': 'area', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Interior': [],
            'Mountain Tower Compass': [],
            'Mountain Tower Treasure': [],
            'Mountain Tower Boss': []}
    },
    'Mountain Tower Compass': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Ascent': []}
    },
    'Mountain Tower Treasure': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Ascent': []}
    },
    'Mountain Tower Boss': {
        'type': 'dungeonboss', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Boss Item': [('item', 'sword'), ('item', 'hammer')]}
    },
    'Mountain Tower Boss Item': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Reward': []}
    },
    'Mountain Tower Reward': {
        'type': 'dungeonreward', 'dungeon': 'Mountain Tower',
        'link': {
            'Mountain Tower Entrance (I)': []}
    },
}



'''
Misery Mire
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Misery Mire Surface': {
        'type': 'area',
        'link': {
            'Desert': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Desert Ridge': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Misery Mire Treasury Entrance (E)': [],
            'Misery Mire Fairy Entrance (E)': [],
            'Desert Cave Valley': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Misery Mire Cave Entrance (E)': [],
            'Misery Mire Dungeon Entrance (E)': [('and', [
                ('medallion', 'Misery Mire'), ('item', 'sword')])]}
    },
    'Misery Mire Treasury Entrance (E)': {
        'type': 'entrance_unique', 'map': 'dark', 'coord': (20, 531),
        'link': {
            'Misery Mire Surface': [],
            'Misery Mire Treasury Entrance (I)': []}
    },
    'Misery Mire Treasury Entrance (I)': {
        'type': 'interior',
        'link': {
            'Misery Mire Treasury Entrance (E)': [],
            'Misery Mire Treasury': [('rabbitbarrier', None)]}
    },
    'Misery Mire Treasury': {
        'type': 'chest', 'map': 'dark', 'coord': (20, 528),
        'link': {
            'Misery Mire Treasury Entrance (I)': []}
    },
    'Misery Mire Fairy Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (66, 531),
        'link': {
            'Misery Mire Surface': [],
            'Misery Mire Fairy Entrance (I)': []}
    },
    'Misery Mire Fairy Entrance (I)': {
        'type': 'interior',
        'link': {
            'Misery Mire Fairy Entrance (E)': []}
    },
    'Misery Mire Cave Entrance (E)': {
        'type': 'entrance', 'map': 'dark', 'coord': (126, 547),
        'link': {
            'Misery Mire Surface': [],
            'Misery Mire Cave Entrance (I)': []}
    },
    'Misery Mire Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Misery Mire Cave Entrance (E)': []}
    },
    'Desert Portal (DW)': {
        'type': 'area',
        'link': {
            'Desert Portal (LW)': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Desert Portal': [('item', 'titansmitts')],
            'Misery Mire Surface': []}
    },

    'Misery Mire Dungeon Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (43, 553),
        'link': {
            'Misery Mire Surface': [],
            'Misery Mire Dungeon Entrance (I)': []}
    },
    'Misery Mire Dungeon Entrance (I)': {
        'type': 'interior',
        'link': {
            'Misery Mire Dungeon Entrance (E)': [],
            'Misery Mire Descent': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'mastersword')]),
                    ('or', [
                        ('item', 'bottle'), ('item', 'bluemail')])])]}
    },
    'Misery Mire Descent': {
        'type': 'area', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Dungeon Entrance (I)': [],
            'Misery Mire Interior': [
                    ('item', 'hookshot'),
                    ('and', [
                        ('item', 'pegasus'),
                        ('settings', 'placement_advanced')])]}
    },

    'Misery Mire Interior': {
        'type': 'area', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Descent': [
                ('item', 'hookshot'), ('item', 'pegasus'), ('item', 'mirror')],
            'Misery Mire Bridge Key': [],
            'Misery Mire Hidden Key': [],
            'Misery Mire Spike Key': [],
            'Misery Mire Map': [('smallkey', ('Misery Mire', 3))],
            'Misery Mire Treasure Room': [
                ('item', 'hookshot'), ('item', 'pegasus')],
            'Misery Mire Upper Key': [],
            'Misery Mire Switch Room': [('smallkey', ('Misery Mire', 3))],
            'Misery Mire Gauntlet Bridge': [('bigkey', 'Misery Mire')]}
    },
    'Misery Mire Bridge Key': {
        'type': 'dungeonchest', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': []}
    },
    'Misery Mire Hidden Key': {
        'type': 'dungeonkey', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': []}
    },
    'Misery Mire Spike Key': {
        'type': 'dungeonchest', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': []}
    },
    'Misery Mire Map': {
        'type': 'dungeonchest', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': []},
    },
    'Misery Mire Treasure Room': {
        'type': 'area', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': [],
            'Misery Mire Treasure': [('bigkey', 'Misery Mire')]}
    },
    'Misery Mire Treasure': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': []},
    },
    'Misery Mire Upper Key': {
        'type': 'dungeonchest', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': []},
    },

    'Misery Mire Switch Room': {
        'type': 'area', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': [],
            'Misery Mire Bari Key': [],
            'Misery Mire Torch Room': [('smallkey', ('Misery Mire', 3))]}
    },
    'Misery Mire Bari Key': {
        'type': 'dungeonkey', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Switch Room': []}
    },
    'Misery Mire Torch Room': {
        'type': 'area', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Switch Room': [],
            'Misery Mire Compass': [('item', 'lantern'), ('item', 'firerod')],
            'Misery Mire Big Key': [('item', 'lantern'), ('item', 'firerod')]}
    },
    'Misery Mire Compass': {
        'type': 'dungeonchest', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Torch Room': [],
            'Misery Mire Interior': []}
    },
    'Misery Mire Big Key': {
        'type': 'dungeonchest', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': []}
    },

    'Misery Mire Gauntlet Bridge': {
        'type': 'area', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Interior': [],
            'Misery Mire Gauntlet': [('item', 'somaria')]}
    },
    'Misery Mire Gauntlet': {
        'type': 'area', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Gauntlet Bridge': [],
            'Misery Mire Boss': [('item', 'bombs')]}
    },
    'Misery Mire Boss': {
        'type': 'dungeonboss', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Boss Item': [('and', [
                ('or', [
                    ('settings', 'placement_advanced'), ('item', 'mastersword'),
                    ('item', 'bow')]),
                ('or', [
                    ('item', 'hammer'), ('item', 'sword'), ('item', 'bow')])])]}
    },
    'Misery Mire Boss Item': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Reward': []},
    },
    'Misery Mire Reward': {
        'type': 'dungeonreward', 'dungeon': 'Misery Mire',
        'link': {
            'Misery Mire Dungeon Entrance (I)': []}
    },
}



'''
Map tracker display
'''

import importlib
import tkinter as tk
import tkinter.ttk as ttk
import typing

from ..config import CONFIG
from ..config.images import image
common = importlib.import_module(
    '..gui-common.maps', package=__package__)
BUTTONTYPE = common.BUTTONTYPE

from . import misc

MAPSCALE = 1

__all__ = 'MapDisplay', 'make_symbol_coordinates'


class MapDisplay(tk.Toplevel):
    '''
    Map window

    Instance variables:
        identifier: map name
        scale: map scale
        imagefile: map image
        entrancetracker: entrance tracker
        other: other (dark/light world) map
        activelink: linked button currently artifically set to active
        button: ocation buttons
        tracker: location trackern
    '''

    def __init__(self, spec: str):
        '''
        Args:
            spec: map type
        '''

        # General initialisation
        assert spec in ('light', 'dark')
        super().__init__()
        self.identifier = spec
        self.scale = CONFIG['map_size']
        self.title('Light World' if spec == 'light' else 'Dark World')
        self.button = {}
        self.entrancetracker = None
        self.other = None
        self.activelink = set()
        self.tracker = None

        # Set up bottom text label.
        self.helpertext = tk.StringVar()
        self.helper = ttk.Label(self, textvariable=self.helpertext)
        self.helper.grid(column=0, row=1, sticky=tk.S)

        # Set background image.
        imagefile = tk.PhotoImage(
            file=image('lightworld' if spec == 'light' else 'darkworld'),
            master=self)
        scaling = MAPSCALE * self.scale
        imgdim = (imagefile.width() * scaling, imagefile.height() * scaling)
        self.map = tk.Canvas(self, height=imgdim[1], width=imgdim[0])
        self.map.grid(column=0, row=0, sticky=misc.A)
        for up in range(1, 1000):
            if not (scaling * up) % 1:
                upscale = int(scaling * up)
                break
        else:
            CONFIG['map_size'] = 1
            self.scale = 1
            assert False
        downscale = int(upscale // scaling)
        if upscale != 1:
            imagefile = imagefile.zoom(upscale, upscale)
        if downscale != 1:
            imagefile = imagefile.subsample(downscale, downscale)
        self.image = self.map.create_image(
            (0, 0), anchor=tk.NW, image=imagefile)
        self.map.bind('<ButtonRelease-3>', lambda _: self._rightclick_button())
        self.imagefile = imagefile

    def add_location_tracker(self, locationtracker) -> None:
        '''
        Add location tracker.

        Args:
            locationtracker: location tracker
        '''

        self.tracker = locationtracker

    def place_buttons(self, locations: dict) -> None:
        '''
        Place buttons on map.

        Args:
            locations: {
               'location name': {'map': str, 'coord': (int, int), 'type': str}}
        '''

        for loc in locations:
            self.add_button(
                loc, locations[loc]['coord'], locations[loc]['type'])

    def add_button(
            self, name: str, coord: typing.Sequence[int],
            buttontype: str) -> None:
        '''
        Add a button.

        Args:
            name: name of new button
            coord: coordinates for centre of button
            buttontype: type of button
        '''

        # Dungeon buttons are different.
        if buttontype == 'dungeon':
            self._add_dungeon_button(name, coord)
            return

        # Make button.
        assert (buttontype in ('chest', 'item', 'cave', 'drop', 'ganon') or
                buttontype.startswith('entrance'))
        displaytext = name
        if buttontype == 'chest':
            buttonfunc = self._chest_icon
        elif buttontype == 'entrance_drop':
            buttonfunc = self._drop_entrance_icon
            displaytext = displaytext.replace(' Entrance (E)', '')
        elif buttontype.startswith('entrance'):
            buttonfunc = self._entrance_icon
            displaytext = displaytext.replace(' Entrance (E)', '')
        elif buttontype == 'item':
            buttonfunc = self._item_icon
        elif buttontype == 'cave':
            buttonfunc = self._cave_icon
        elif buttontype == 'drop':
            buttonfunc = self._drop_icon
        elif buttontype == 'ganon':
            buttonfunc = self._ganon_icon
        else:
            buttonfunc = self._item_icon
        new = buttonfunc(coord)
        self.button[name] = {
            'id': new, 'type': buttontype, 'display': displaytext,
            'available': 'unavailable', 'entrance': False}
        if name not in self.tracker:
            self.tracker[name] = True

        # Assign button functions.
        self.map.tag_bind(
            new, '<Enter>', lambda _: self._hover(name))
        self.map.tag_bind(
            new, '<Leave>', lambda _: self._unhover(name))
        self.map.tag_bind(
            new, '<ButtonRelease-1>', lambda _: self._leftclick_button(name))
        if buttontype.startswith('entrance'):
            self.map.tag_bind(
                new, '<ButtonRelease-2>', lambda _: self._middleclick_button(name))
            self.map.tag_bind(
                new, '<ButtonRelease-3>', lambda _: self._rightclick_button(name))

    def _hover(self, button: str) -> None:
        '''
        Set bottom display.

        Args:
            button: button under mouse
        '''

        displaytext = self.button[button]['display']
        if self.button[button]['type'].startswith('entrance'):

            # Entrance leading to itself.
            try:
                same = self.entrancetracker[button] == button.replace(
                    ' (E)', ' (I)')
            except KeyError:
                same = False
            if button in self.entrancetracker and same:
                self.set_entrance_colour(button, 'default')

            # Entrances reachable from entrance.
            linked = self.entrancetracker.find_remote(button)
            for loc in linked:
                if loc in self.button:
                    self.set_entrance_colour(loc, 'linked')
                else:
                    self.other.set_entrance_colour(loc, 'linked')

            # Entrance leading to interior.
            if button in self.entrancetracker and not same:
                linked = self.entrancetracker[button].replace(
                    ' Entrance (I)', '')
                displaytext = '{0:s} → {1:s}'.format(displaytext, linked)
                if displaytext != linked:
                    linkbutton = '{0:s} Entrance (E)'.format(linked)
                    if linkbutton in self.button:
                        self.set_entrance_colour(linkbutton, 'interior')
                    else:
                        self.other.set_entrance_colour(linkbutton, 'interior')

            # Interior reached via entrance.
            interior = button.replace(' (E)', ' (I)')
            if interior in self.entrancetracker and not same:
                linked = self.entrancetracker[interior]
                if displaytext != linked.replace(' Entrance (E)', ''):
                    if linked in self.button:
                        self.set_entrance_colour(linked, 'entrance')
                    else:
                        self.other.set_entrance_colour(linked, 'entrance')

        self.helpertext.set(displaytext)

    def _unhover(self, button: str) -> None:
        '''
        Set bottom display.

        Args:
            button: button under mouse
        '''

        self.helpertext.set('')
        while self.activelink:
            self._set_colour(self.activelink.pop())
        while self.other.activelink:
            self.other._set_colour(self.other.activelink.pop())

    def _leftclick_button(self, name: str) -> None:
        '''
        Event on left-clicking a button.

        Args:
            name: name of clicked-on button
        '''

        self.tracker[name] = not self.tracker[name]
        self._set_colour(name)
        self.tracker.save()

    def _middleclick_button(self, name: str) -> None:
        '''
        Event on middle-clicking a button.

        Args:
            name: name of clicked on button
        '''

        new = self.entrancetracker.event(name)
        if not new and self.entrancetracker.armed:
            self.helper.configure(background='yellow')
        else:
            self.helper.configure(background='')
        if new:
            self.set_entrance()

    def _rightclick_button(self, name: str = None) -> None:
        '''
        Event on right-clicking a button.

        Args:
            name: name of clicked on button
        '''

        self.entrancetracker.abort(name)
        self.helper.configure(background='')
        if name:
            self.set_entrance()
            self._hover(name)

    def _add_dungeon_button(
            self, name: str, coord: typing.Sequence[int]) -> None:
        '''
        Add dungeon button.

        Args:
            name: name of dungeon
            coord: location on map
        '''

        shape = -13, -13, 13, 13
        loc = make_symbol_coordinates(coord, shape, self.scale)
        new_outer = self.map.create_rectangle(*loc)
        shape = -7, -7, 7, 7
        loc = make_symbol_coordinates(coord, shape, self.scale)
        new_inner = self.map.create_rectangle(*loc)
        self.button[name] = {
            'id': (new_outer, new_inner), 'type': 'dungeon', 'display': name,
            'available': ['unavailable', 'unavailable']}
        if name not in self.tracker:
            self.tracker[name] = [True, True]
        
        for new in (new_outer, new_inner):
            self.map.tag_bind(
                new, '<Enter>', lambda _: self.helpertext.set(name))
            self.map.tag_bind(
                new, '<Leave>', lambda _: self.helpertext.set(''))

    def remove_buttons(self) -> None:
        '''
        Remove all buttons from map.
        '''

        for button in self.button.values():
            if button['type'] == 'dungeon':
                self.map.delete(button['id'][0])
                self.map.delete(button['id'][1])
            else:
                self.map.delete(button['id'])
        self.button = {}

    def set_availability(self, name: str, availability: str) -> None:
        '''
        Set whether given location is available.

        Args:
            name: location name
            availability: 'available', 'unavailable', 'visible'
        '''

        assert availability in ('available', 'unavailable', 'visible')
        self.button[name]['available'] = availability
        self._set_colour(name)

    def set_dungeon_availability(
            self, name: str, availability: (str, str)) -> None:
        '''
        Set whether given dungeon is available.

        Args:
            name: location name
            availability: two of 'available', 'unavailable', 'visible'
        '''

        for idx, avail in enumerate(availability):
            assert avail in ('available', 'unavailable', 'visible')
            self.button[name]['available'][idx] = avail
        self._set_colour(name)
        self.tracker.save()

    def set_entrance(self) -> None:
        '''
        Update entrance display.
        '''

        for button in self.button:
            if self.button[button]['type'] == 'dungeon':
                continue
            if (button in self.entrancetracker and
                not self.button[button]['entrance']):
                self.button[button]['entrance'] = True
                self.entrancetracker.update()
            elif (button not in self.entrancetracker and
                  self.button[button]['entrance']):
                self.button[button]['entrance'] = False
                self.entrancetracker.update()
        for button in self.other.button:
            if self.other.button[button]['type'] == 'dungeon':
                continue
            if (button in self.entrancetracker and
                not self.other.button[button]['entrance']):
                self.other.button[button]['entrance'] = True
                self.entrancetracker.update()
            elif (button not in self.entrancetracker and
                  self.other.button[button]['entrance']):
                self.other.button[button]['entrance'] = False
                self.entrancetracker.update()

    def _set_colour(self, button: str) -> None:
        '''
        Set button colour.

        Args:
            button: button name
        '''

        if self.button[button]['type'] == 'dungeon':
            self._set_dungeon_colour(button)
            return
        if not self.tracker[button]:
            colour = 'checked'
        else:
            colour = self.button[button]['available']
        if self.button[button]['type'] in BUTTONTYPE:
            colours = BUTTONTYPE[self.button[button]['type']]['colours']
        else:
            colours = BUTTONTYPE['standard']['colours']
        if self.button[button]['entrance']:
            activeoutline = 'black'
            activewidth = 3
            outline = 3
        else:
            activeoutline = 'black'
            activewidth = 1
            outline = 1
        self.map.itemconfigure(
            self.button[button]['id'],
            activefill=colours[colour]['active'],
            activeoutline=activeoutline,
            activewidth=activewidth,
            fill=colours[colour]['normal'],
            outline='black', width=outline)

    def set_entrance_colour(self, button: str, linktype: str) -> None:
        '''
        Set colour of linked entrance button.

        Args:
            button: button name
            linktime: 'entrance', 'interior', 'linked'
        '''

        assert linktype in ('entrance', 'interior', 'linked', 'default')
        if not self.tracker[button]:
            colour = 'checked'
        else:
            colour = self.button[button]['available']
        if self.button[button]['type'] in BUTTONTYPE:
            colours = BUTTONTYPE[self.button[button]['type']]['colours']
        else:
            colours = BUTTONTYPE['standard']['colours']
        if button in self.activelink:
            if linktype == 'interior':
                linkcolour = 'magenta'
            elif linktype == 'entrance':
                linkcolour = 'cyan'
            else:
                linkcolour = 'orange'
            self.map.itemconfigure(
                self.button[button]['id'], fill=colours[colour]['active'],
                outline=linkcolour, width=3)
        elif linktype == 'entrance':
            self.map.itemconfigure(
                self.button[button]['id'], fill=colours[colour]['active'],
                outline='yellow', width=3)
        elif linktype == 'interior':
            self.map.itemconfigure(
                self.button[button]['id'], fill=colours[colour]['active'],
                outline='red', width=3)
        elif linktype == 'linked':
            self.map.itemconfigure(
                self.button[button]['id'], fill=colours[colour]['active'],
                outline='blue', width=3)
        elif linktype == 'default':
            self.map.itemconfigure(
                self.button[button]['id'], activefill=colours[colour]['active'],
                activeoutline='white', activewidth=3)
        self.activelink.add(button)

    def _set_dungeon_colour(self, button: str) -> None:
        '''
        Set colour of dungeon button.

        Args:
            button: dungeon name
        '''

        for bpart in range(2):
            if not self.tracker[button][bpart]:
                colour = 'checked'
            else:
                colour = self.button[button]['available'][bpart]
            colours = BUTTONTYPE['standard']['colours']
            self.map.itemconfigure(
                self.button[button]['id'][bpart],
                activefill=colours[colour]['normal'],
                fill=colours[colour]['normal'],
                outline='black', width=1)

    def reset(self) -> None:
        '''
        Reset state of all buttons.
        '''

        self.entrancetracker.reset()
        self.tracker.reset()
        self.set_entrance()
        for button in self.button:
            if self.button[button]['type'] == 'dungeon':
                self.tracker[button] = [True, True]
            else:
                self.tracker[button] = True
            self._set_colour(button)

    def add_entrance_tracker(self, entrancetracker) -> None:
        '''
        Add entrance tracker.
        '''

        self.entrancetracker = entrancetracker

    def _chest_icon(self, location: typing.Sequence[int]) -> int:
        shape = -7, -7, 7, 7
        loc = make_symbol_coordinates(location, shape, self.scale)
        new = self.map.create_rectangle(*loc)
        return new

    def _entrance_icon(self, location: typing.Sequence[int]) -> int:
        shape = -5, -5, 5, 5
        loc = make_symbol_coordinates(location, shape, self.scale)
        new = self.map.create_rectangle(*loc)
        return new

    def _item_icon(self, location: typing.Sequence[int]) -> int:
        shape = -7, -7, 7, 7
        loc = make_symbol_coordinates(location, shape, self.scale)
        new = self.map.create_oval(*loc)
        return new

    def _cave_icon(self, location: typing.Sequence[int]) -> int:
        shape = -7, 7, 0, -7, 7, 7
        loc = make_symbol_coordinates(location, shape, self.scale)
        new = self.map.create_polygon(*loc)
        return new

    def _drop_icon(self, location: typing.Sequence[int]) -> int:
        shape = -7, -7, 0, 7, 7, -7
        loc = make_symbol_coordinates(location, shape, self.scale)
        new = self.map.create_polygon(*loc)
        return new

    def _drop_entrance_icon(self, location: typing.Sequence[int]) -> int:
        shape = -6, -6, 0, 6, 6, -6
        loc = make_symbol_coordinates(location, shape, self.scale)
        new = self.map.create_polygon(*loc)
        return new

    def _ganon_icon(self, location: typing.Sequence[int]) -> int:
        '''Go-mode symbol'''
        shape = (0, -20, 5, -5, 20, -5, 8, 5, 12, 20, 0, 11,
                 -12, 20, -8, 5, -20, -5, -5, -5)
        loc = make_symbol_coordinates(location, shape, self.scale)
        new = self.map.create_polygon(*loc)
        return new


def make_symbol_coordinates(
        location: typing.Sequence[int], shape: typing.Sequence[int],
        mapscale: float) -> list:
    '''
    Create corner points for map symbol.

    Args:
        location: centre of symbol
        shape: symbol corners relative to centre point
        mapscale: additional map scaling factor
    Returns:
        list: flat list of coordinates for symbol creation
    '''

    loc = list(location[:2]) * (len(shape) // 2)
    scaled = tuple(int(c * MAPSCALE * mapscale) for c in shape)
    loc = [loc[l] * mapscale + scaled[l] for l in range(len(scaled))]
    return loc




'''
Item management
'''

import configparser
import importlib
import os.path
import json

from ..config.images import image
layoutstorage = importlib.import_module(
    '..gui-common.items', package=__package__)

from .itemobj import ItemObj
from .lists import ITEMS
from . import storage

__all__ = 'ItemTracker',


def get_layout() -> dict:
    '''
    Load (or create) item layout.

    Returns:
        dict: item layout in format {identifier: (column, row)}
    '''

    try:
        layout = layoutstorage.load_items()
    except (layoutstorage.NoConfig, configparser.Error):
        layout = {}
        for item in ITEMS:
            layout[item[0]] = item[1]
        layoutstorage.new_item(layout)
        layout = layoutstorage.load_items()
    return layout


class ItemTracker(dict):
    '''
    Inventory item tracker.
    '''

    def __init__(self, worldtracker):
        '''
        Args:
            worldtracker: world state tracker
        '''

        layout = get_layout()

        super().__init__()
        delayed_link = []
        for rawitem in ITEMS:
            item = ItemObj(
                rawitem[0], rawitem[1], rawitem[2], tuple(rawitem[3].keys()),
                tuple(rawitem[3].values()),
                rawitem[4]['links'] if 'links' in rawitem[4] else {},
                rawitem[4]['default'] if 'default' in rawitem[4] else 0,
                worldtracker)
            try:
                item.location = layout[item.identifier.lower()]
            except KeyError:
                pass
            self[item.identifier] = item
        for item in self:
            for linkitem in self[item].link:
                self[item].linkitems[linkitem] = self[linkitem]
        data = storage.load_items()
        self.restore(data)

    def reset(self) -> None:
        '''
        Reset all items.
        '''

        for item in self:
            self[item].reset()

    def save(self) -> dict:
        '''
        Save current item setup.
        '''

        storage.store_items(self.store())

    def store(self) -> dict:
        '''
        Return current item setup info for storage.

        Returns:
            inventory: item setup info
        '''

        inventory = {}
        for item in self:
            inventory[item] = self[item].inventory
        return inventory

    def restore(self, inventory) -> None:
        '''
        Restore current item setup from file.

        Args:
            inventory: information from file
        '''

        for item in inventory:
            if item in self:
                self[item].restore_inventory(inventory[item])



'''
Help window
'''

import importlib
import tkinter as tk
import tkinter.ttk as ttk

from ..config import CONFIG
common = importlib.import_module(
    '..gui-common.maps', package=__package__)
BUTTONTYPE = common.BUTTONTYPE

from .maps import make_symbol_coordinates
from . import misc

__all__ = 'HelpWindow',


class HelpWindow(tk.Toplevel):
    '''
    Help window
    '''

    def __init__(self):
        super().__init__()

        self.title('Help')
        self.frame = ttk.Frame(self)
        self.frame.grid(sticky=misc.A)
        self.widgets = []

        self.leftframe = ttk.Frame(self)
        self.leftframe.grid(column=0, row=0, sticky=misc.A)
        self.middleframe = ttk.Frame(self)
        self.middleframe.grid(column=1, row=0, sticky=misc.A)
        self.rightframe = ttk.Frame(self)
        self.rightframe.grid(column=2, row=0, sticky=misc.A)

        self._item_buttons()
        self._dungeon_buttons()
        self._entrance_buttons()
        self._map_control()

    def _item_buttons(self) -> None:
        '''
        Item colours
        '''

        widget = ttk.LabelFrame(self.leftframe, text='Item locations')
        widget.grid(column=0, row=0, sticky=tk.N+tk.E+tk.W)
        shape = -7, -7, 7, 7
        loc = make_symbol_coordinates((16, 16), shape, 1)

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=0, sticky=misc.A)
        button = canvas.create_oval(*loc)
        canvas.itemconfigure(
            button, 
            activefill=BUTTONTYPE['standard']['colours']['available']['active'],
            fill=BUTTONTYPE['standard']['colours']['available']['normal'])
        text = ttk.Label(widget, text='Available')
        text.grid(column=1, row=0, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=1, sticky=misc.A)
        button = canvas.create_oval(*loc)
        canvas.itemconfigure(
            button, 
            activefill=BUTTONTYPE['standard']['colours']['checked']['active'],
            fill=BUTTONTYPE['standard']['colours']['checked']['normal'])
        text = ttk.Label(widget, text='Already checked')
        text.grid(column=1, row=1, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=2, sticky=misc.A)
        button = canvas.create_oval(*loc)
        canvas.itemconfigure(
            button, 
            activefill=BUTTONTYPE[
                'standard']['colours']['unavailable']['active'],
            fill=BUTTONTYPE['standard']['colours']['unavailable']['normal'])
        text = ttk.Label(widget, text='Unavailable')
        text.grid(column=1, row=2, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=3, sticky=misc.A)
        button = canvas.create_oval(*loc)
        canvas.itemconfigure(
            button, 
            activefill=BUTTONTYPE['standard']['colours']['visible']['active'],
            fill=BUTTONTYPE['standard']['colours']['visible']['normal'])
        text = ttk.Label(widget, text='Unavailable, but visible')
        text.grid(column=1, row=3, sticky=tk.W)
        self.widgets.extend((canvas, widget))

    def _dungeon_buttons(self) -> None:
        '''
        Dungeon colours
        '''

        widget = ttk.LabelFrame(self.leftframe, text='Dungeon locations')
        widget.grid(column=0, row=1, sticky=tk.N+tk.E+tk.W)
        shape = -13, -13, 13, 13
        loc = make_symbol_coordinates((16, 16), shape, 1)

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=0, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['available']['normal'])
        text = ttk.Label(widget, text='Entirely available')
        text.grid(column=1, row=0, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=1, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['checked']['normal'])
        text = ttk.Label(widget, text='Fully checked')
        text.grid(column=1, row=1, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=2, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['unavailable']['normal'])
        text = ttk.Label(widget, text='All unavailable')
        text.grid(column=1, row=2, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=3, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['visible']['normal'])
        text = ttk.Label(widget, text='Possibly or partially available')
        text.grid(column=1, row=3, sticky=tk.W)
        self.widgets.extend((canvas, widget))

    def _entrance_buttons(self):
        '''
        Entrance button colours
        '''

        widget = ttk.LabelFrame(self.middleframe, text='Entrances')
        widget.grid(column=0, row=0, sticky=tk.N)
        shape = -5, -5, 5, 5
        loc = make_symbol_coordinates((16, 16), shape, 1)

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=0, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['available']['active'],
            outline='white', width=3)
        text = ttk.Label(widget, text='Entrance leads to default location')
        text.grid(column=1, row=0, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=1, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['available']['active'],
            outline='red', width=3)
        text = ttk.Label(widget, text='Selected entrance leads here')
        text.grid(column=1, row=1, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=2, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['available']['active'],
            outline='yellow', width=3)
        text = ttk.Label(
            widget, text='This entrance leads to selected location')
        text.grid(column=1, row=2, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=3, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['available']['active'],
            outline='orange', width=3)
        text = ttk.Label(
            widget,
            text='Entrances are switched between this and selected location')
        text.grid(column=1, row=3, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=4, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['available']['active'],
            outline='blue', width=3)
        text = ttk.Label(
            widget,
            text='This entrance leads to same interior as selected one.')
        text.grid(column=1, row=4, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=5, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['available']['active'],
            outline='magenta', width=3)
        text = ttk.Label(
            widget,
            text='Combination of red and blue.')
        text.grid(column=1, row=5, sticky=tk.W)
        self.widgets.extend((canvas, widget))

        canvas = tk.Canvas(widget, height=32, width=32)
        canvas.grid(column=0, row=6, sticky=misc.A)
        button = canvas.create_rectangle(*loc)
        canvas.itemconfigure(
            button,
            fill=BUTTONTYPE['standard']['colours']['available']['active'],
            outline='cyan', width=3)
        text = ttk.Label(
            widget,
            text='Combination of yellow and blue.')
        text.grid(column=1, row=6, sticky=tk.W)
        self.widgets.extend((canvas, widget))

    def _map_control(self) -> None:
        '''
        Map controls
        '''

        widget = ttk.LabelFrame(self.rightframe, text='Entrance map control')
        widget.grid(column=0, row=0, sticky=tk.N)

        button = ttk.Label(widget, text='Left-click:')
        button.grid(column=0, row=0, sticky=misc.A)
        spacer = ttk.Label(widget, width=1)
        spacer.grid(column=1, row=0, sticky=misc.A)
        text = ttk.Label(widget, text='Mark and unmark locations as checked')
        text.grid(column=2, row=0, sticky=tk.W)

        button = ttk.Label(widget, text='Middle-click:')
        button.grid(column=0, row=1, sticky=misc.A)
        spacer = ttk.Label(widget, width=1)
        spacer.grid(column=1, row=1, sticky=misc.A)
        text = ttk.Label(widget, text='Link entrance to location')
        text.grid(column=2, row=1, sticky=tk.W)

        button = ttk.Label(widget, text='Right-click:')
        button.grid(column=0, row=2, sticky=misc.A)
        spacer = ttk.Label(widget, width=1)
        spacer.grid(column=1, row=2, sticky=misc.A)
        text = ttk.Label(widget, text='Abort linking or unlink location')
        text.grid(column=2, row=2, sticky=tk.W)

    def reset(self) -> None:
        '''
        Dummy method
        '''

        pass


def start(*args) -> tk.Toplevel:
    '''
    Open help window.

    Returns:
        tk.Toplevel: help window
    '''

    return HelpWindow()



'''
Northern Light World
'''

__all__ = 'LOCATIONS',


LOCATIONS= {
    'Lost Woods': {
        'type': 'area',
        'link': {
            'Skull Woods': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Hidden Skull Woods': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Light Bottle Ocarina': [],
            'Kakariko': [('rabbitbarrier', None)],
            'Kakariko Portal (LW)': [('item', 'hammer')],
            'North Light World': [],
            'Master Sword Pedestal': [('and', [
                ('pendant', 'courage'),
                ('pendant', 'power'),
                ('pendant', 'wisdom'),
                ('or', [
                    ('settings', 'placement_advanced'), ('item', 'mudora')])])],
            'Mushroom': [('rabbitbarrier', None)],
            "Thief's Cave Drop Entrance (E)": [('rabbitbarrier', None)],
            "Thief's Cave Entrance Entrance (E)": [],
            'Northern Chest Game Entrance (E)': []},
        'visible': {
            'Master Sword Pedestal': [('item', 'mudora')]}
    },
    'Master Sword Pedestal': {
        'type': 'item', 'map': 'light', 'coord': (27, 33),
        'link': {
            'Lost Woods': []}
    },
    'Mushroom': {
        'type': 'item', 'map': 'light', 'coord': (79, 57),
        'link': {
            'Lost Woods': [('rabbitbarrier', None)]}
    },
    "Thief's Cave Drop Entrance (E)": {
        'type': 'entrance_drop', 'map': 'light', 'coord': (125, 84),
        'link': {
            "Thief's Cave Drop Entrance (I)": []}
    },
    "Thief's Cave Drop Entrance (I)": {
        'type': 'interior',
        'link': {
            "Thief's Cave": []}
    },
    "Thief's Cave": {
        'type': 'drop', 'map': 'light', 'coord': (125, 87),
        'link': {
            "Thief's Cave Entrance": []}
    },
    "Thief's Cave Entrance": {
        'type': 'area',
        'link': {
            "Thief's Cave Entrance Entrance (I)": []},
        'visible': {"Thief's Cave": []}
    },
    "Thief's Cave Entrance Entrance (I)": {
        'type': 'interior',
        'link': {
            "Thief's Cave Entrance": [],
            "Thief's Cave Entrance Entrance (I)": []}
    },
    "Thief's Cave Entrance Entrance (E)": {
        'type': 'entrance_unique', 'map': 'light', 'coord': (122, 99),
        'link': {
            "Thief's Cave Entrance": [],
            'Lost Woods': []}
    },
    'Northern Chest Game Entrance (E)': {
        'type': 'entrance', 'map': 'light', 'coord': (123, 10),
        'link': {
            'Lost Woods': [],
            'Northern Chest Game Entrance (I)': []}
    },
    'Northern Chest Game Entrance (I)': {
        'type': 'interior',
        'link': {
            'Northern Chest Game Entrance (E)': []}
    },


    'North Light World': {
        'type': 'area',
        'link': {
            'North Dark World': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Kakariko': [],
            'Lost Woods': [],
            'Kakariko Fortune Teller Entrance (E)': [],
            'Hyrule Castle': [],
            'Crossworld Waterway': [('item', 'flippers')],
            "Lumberjacks' House Entrance (E)": [],
            'Cut Tree Drop Entrance (E)': [('and', [
                ('access', 'Castle Tower Reward'), ('item', 'pegasus')])],
            'Cut Tree Cave Entrance (E)': [],
            'Hebra Ascent (Bottom) Entrance (E)': [('item', 'powerglove')],
            'Pegasus Cave Entrance (E)': [('item', 'pegasus')],
            'Sanctuary Entrance Entrance (E)': [],
            'Sanctuary Drop Entrance (E)': [('item', 'powerglove')],
            'Graveyard Cave Entrance': [
                ('settings', 'inverted'),
                ('and', [
                    ('glitch', 'overworld'), ('item', 'pegasus')])],
            "King's Tomb Entrance (E)": [('and', [
                ('item', 'titansmitts'), ('item', 'pegasus')])],
            'North Fairy Pond Drop Entrance (E)': [('rabbitbarrier', None)],
            'North Fairy Pond Cave Entrance (E)': [],
            'Witch Hut River': [('item', 'flippers')],
            'Little River': [('item', 'flippers')],
            'East Light World': [],
            'West Lower Mount Hebra': [
                ('glitch', 'major'),
                ('and', [
                    ('glitch', 'overworld'), ('item', 'pegasus')])]}
    },
    'Kakariko Fortune Teller Entrance (E)': {
        'type': 'entrance', 'map': 'light', 'coord': (125, 214),
        'link': {
            'North Light World': [],
            'Kakariko Fortune Teller Entrance (I)': []}
    },
    'Kakariko Fortune Teller Entrance (I)': {
        'type': 'interior',
        'link': {
            'Kakariko Fortune Teller Entrance (E)': []}
    },
    'Crossworld Waterway': {
        'type': 'area',
        'link': {
            'Lake Hylia': [],
            'North Light World': []}
    },
    "Lumberjacks' House Entrance (E)": {
        'type': 'entrance', 'map': 'light', 'coord': (223, 40),
        'link': {
            'North Light World': [],
            "Lumberjacks' House Entrance (I)": []}
    },
    "Lumberjacks' House Entrance (I)": {
        'type': 'interior',
        'link': {
            "Lumberjacks' House Entrance (E)": []}
    },
    'Cut Tree Drop Entrance (E)': {
        'type': 'entrance_drop', 'map': 'light', 'coord': (200, 49),
        'link': {
            'Cut Tree Drop Entrance (I)': []}
    },
    'Cut Tree Drop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Cut Tree': []}
    },
    'Cut Tree': {
        'type': 'drop', 'map': 'light', 'coord': (200, 49),
        'link': {
            'Cut Tree Viewpoint': []}
    },
    'Cut Tree Cave Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (220, 21),
        'link': {
            'North Light World': [],
            'Cut Tree Cave Entrance (I)': []}
    },
    'Cut Tree Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Cut Tree Cave Entrance (E)': [],
            'Cut Tree Viewpoint': []}
    },
    'Cut Tree Viewpoint': {
        'type': 'area',
        'link': {
            'Cut Tree Cave Entrance (I)': []},
        'visible': {'Cut Tree': []}
    },
    'Hebra Ascent (Bottom) Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (236, 116),
        'link': {
            'North Light World': [],
            'Hebra Ascent (Bottom) Entrance (I)': [('nosettings', 'inverted')],
            'Bumper Cave Bottom Entrance Entrance (I)': [
                ('settings', 'inverted')]}
    },
    'Hebra Ascent (Bottom) Entrance (I)': {
        'type': 'interior',
        'link': {
            'Hebra Ascent (Bottom) Entrance (E)': [('nosettings', 'inverted')],
            'Bumper Cave Bottom Entrance Entrance (E)': [
                ('settings', 'inverted')],
            'Hebra Ascent': []}
    },
    'Hebra Descent (Bottom) Entrance (I)': {
        'type': 'interior',
        'link': {
            'Hebra Descent': [],
            'Hebra Descent (Bottom) Entrance (E)': [('nosettings', 'inverted')],
            'Bumper Cave Bottom Entrance Entrance (E)': [
                ('settings', 'inverted')]}
    },
    'Hebra Descent (Bottom) Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (238, 101),
        'link': {
            'Hebra Descent (Bottom) Entrance (I)': [('nosettings', 'inverted')],
            'Bumper Cave Top Entrance Entrance (I)': [
                ('settings', 'inverted')],
            'Hebra Descent End': []}
    },
    'Hebra Descent End': {
        'type': 'area',
        'link': {
            'Bumper Cave Top': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'Hebra Descent (Bottom) Entrance (E)': [],
            'North Light World': []}
    },
    'Pegasus Cave Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (259, 194),
        'link': {
            'North Light World': [],
            'Pegasus Cave Entrance (I)': []}
    },
    'Pegasus Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Pegasus Cave Entrance (E)': [],
            'Pegasus Cave': []}
    },
    'Pegasus Cave': {
        'type': 'drop', 'map': 'light', 'coord': (259, 194),
        'link': {
            'Pegasus Cave Entrance (I)': []}
    },
    'Sanctuary Entrance Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (306, 176),
        'link': {
            'North Light World': [],
            'Sanctuary Entrance Entrance (I)': []}
    },
    'Sanctuary Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Sanctuary Entrance Entrance (E)': [],
            'Sanctuary': [('rabbitbarrier', None)]}
    },
    'Sanctuary Drop Entrance (E)': {
        'type': 'entrance_drop', 'map': 'light', 'coord': (345, 194),
        'link': {
            'Sanctuary Drop Entrance (I)': []}
    },
    'Sanctuary Drop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Sewers Last Hall': []}
    },
    'Graveyard Cave Entrance': {
        'type': 'area',
        'link': {
            'North Light World': [],
            'Graveyard Cave Entrance (E)': []}
    },
    'Graveyard Cave Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (378, 181),
        'link': {
            'Graveyard Cave Entrance': [],
            'Graveyard Cave Entrance (I)': []}
    },
    'Graveyard Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'Graveyard Cave Entrance (E)': [],
            'Graveyard Cave Front': []}
    },
    'Graveyard Cave Front': {
        'type': 'area',
        'link': {
            'Graveyard Cave Entrance (I)': [],
            'Graveyard Cave': [('item', 'bombs')]}
    },
    'Graveyard Cave': {
        'type': 'cave', 'map': 'light', 'coord': (378, 181),
        'link': {
            'Graveyard Cave Front': []}
    },
    "King's Tomb Entrance (E)": {
        'type': 'entrance_unique', 'map': 'light', 'coord': (399, 197),
        'link': {
            'North Light World': [('item', 'titansmitts')],
            "King's Tomb Entrance (I)": []}
    },
    "King's Tomb Entrance (I)": {
        'type': 'interior',
        'link': {
            "King's Tomb Entrance (E)": [],
            "King's Tomb": []}
    },
    "King's Tomb": {
        'type': 'drop', 'map': 'light', 'coord': (399, 197),
        'link': {
            "King's Tomb Entrance (I)": []}
    },
    'North Fairy Pond Drop Entrance (E)': {
        'type': 'entrance_drop', 'map': 'light', 'coord': (426, 206),
        'link': {
            'North Fairy Pond Drop Entrance (I)': []}
    },
    'North Fairy Pond Drop Entrance (I)': {
        'type': 'interior',
        'link': {
            'North Fairy Pond': []}
    },
    'North Fairy Pond': {
        'type': 'area',
        'link': {
            'North Fairy Pond Cave': []}
    },
    'North Fairy Pond Cave Entrance (E)': {
        'type': 'entrance_unique', 'map': 'light', 'coord': (443, 181),
        'link': {
            'North Light World': [],
            'North Fairy Pond Cave Entrance (I)': []}
    },
    'North Fairy Pond Cave Entrance (I)': {
        'type': 'interior',
        'link': {
            'North Fairy Pond Cave Entrance (E)': [],
            'North Fairy Pond Cave': []}
    },
    'North Fairy Pond Cave': {
        'type': 'area',
        'link': {
            'North Fairy Pond Cave Entrance (I)': []}
    },
    'Little River': {
        'type': 'area',
        'link': {
            'Dark River': [('and', [
                ('settings', 'inverted'), ('item', 'mirror')])],
            'North Light World': []}
    },
}



'''
Rules management
'''

import importlib
import typing

LOCATIONIMPORTS = (
    'castle', 'darkpalace', 'darkworld_east', 'darkworld_north',
    'darkworld_south', 'deathmountain', 'desert', 'dungeons', 'ganonstower',
    'hebra', 'icepalace', 'kakariko', 'lightworld_east', 'lightworld_north',
    'lightworld_south', 'miserymire', 'mountaintower', 'ocarina', 'swamppalace',
    'skullwoods', 'thievestown', 'turtlerock')
LOCATIONS = {}
for locimp in LOCATIONIMPORTS:
    addloc = importlib.import_module(
        '.{0:s}'.format(locimp), package=__package__)
    LOCATIONS.update(addloc.LOCATIONS)
del addloc, locimp

__all__ = 'Ruleset',


class Ruleset(dict):
    '''
    Placement ruleset
    '''

    def __init__(self):
        super().__init__()
        self.dungeons = {}
        self.parse_locations()

    def parse_locations(self) -> None:
        '''
        Create location database.
        '''

        for loc in LOCATIONS:
            self[loc] = Location(LOCATIONS[loc])
            if 'dungeon' in LOCATIONS[loc]:
                if self[loc].dungeon not in self.dungeons:
                    self.dungeons[self[loc].dungeon] = Dungeon()
                self.dungeons[self[loc].dungeon][loc] = self[loc]

    def locations(self, gametype: str, maptype: str) -> dict:
        '''
        Return a list of locations to be displayed on a map.

        Args:
            gametype: 'item' or 'entrance'
            maptype: 'light' or 'dark'
        Returns:
            dict: {
                'location name': {'map': str, 'coord': (int, int), 'type': str}}
        '''

        assert gametype in ('item', 'item_retro', 'entrance')
        assert maptype in ('light', 'dark')
        if gametype == 'item':
            valid = 'chest', 'item', 'cave', 'drop', 'dungeon', 'ganon'
        elif gametype == 'item_retro':
            valid = (
                'chest', 'item', 'cave', 'drop', 'dungeon', 'ganon', 'entrance',
                'entrance_shop')
        else:
            valid = ('entrance', 'entrance_unique', 'entrance_shop',
                     'entrance_drop', 'entrance_dungeon', 'item', 'dungeon',
                     'ganon')
        ret = {}
        for loc in self:
            if self[loc].type in valid and self[loc].map == maptype:
                ret[loc] = {
                    'map': self[loc].map, 'coord': self[loc].coord,
                    'type': self[loc].type}
        return ret

    def disconnect_entrances(self) -> None:
        '''
        Disconnect all entrances from each other.
        '''

        for loc in self:
            for link in tuple(self[loc].link.keys()):
                if ((self[loc].type.startswith('entrance') and
                     self[link].type == 'interior') or
                    (self[loc].type == 'interior' and
                     self[link].type.startswith('entrance'))):
                    del self[loc].link[link]


class Location(object):
    '''
    Location object

    Instance variables:
        type:
            - 'chest': chest in house or in the open
            - 'interior': interior connector; always has 'entrance' counterpart
            - 'entrance': exterior connector shuffable by entrance randomiser
            - 'entrance_unique': like entrance, but no take-any cave
            - 'area': generic area linking other locations
            - 'item': item in the open (or semi-hidden, e.g. buried)
            - 'cave': open or chested item in a cave
            - 'drop': drophole or downstairs cave
            - 'dungeonchest': generic randomised item found in dungeon
            - 'dungeonchest_nokey': can't hold a key
            - 'dungeonkey': fixed small key found in dungeon
            - 'dungeonboss': dungeon boss
            - 'dungeonreward': pendant or crystal
            - 'dungeon': the overall dungeon itself
            - 'ganon': only used for the final bossfight
        map: map on which location is displayed (not for 'area' or 'interior')
        coord: coordinates of location on map (not for 'area' or 'interior')
        dungeon: parent dungeon (only for dungeon locations)
        link: locations to which this location connects, following the format
            {'linked location': {'requirement type': 'requirement', ...}};
            possible requirements are:
            - 'settings': required settings
            - 'nosettings': only if given setting is not active
            - 'item': item in possession
            - 'access': name of another location which needs to be available
            - 'medallion': dynamic medallion requirement for two dungeons
            - 'pendant': pendant in posession
            - 'crystals': certain crystals in posession
            - 'smallkeys': given number of dungeon chests being available
            - 'bigkey': all dungeon chests being available
            - 'macro': more complex requirements
        visible: list of locations visible from here
        checked: boolean always set to False or None
    '''

    def __init__(self, location: dict):
        '''
        Args:
            location: location info; if None then object is uninitialised
        '''

        assert location['type'] in (
            'chest', 'interior', 'entrance', 'entrance_unique', 'entrance_shop',
            'entrance_drop', 'entrance_dungeon', 'area', 'item', 'cave', 'drop',
            'dungeonchest', 'dungeonchest_nokey', 'dungeonkey', 'dungeonboss',
            'dungeonreward', 'dungeon', 'ganon')
        self.type = location['type']
        if (
                self.type not in ('area', 'interior') and
                not self.type.startswith('dungeon')):
            self.map = location['map']
            self.coord = location['coord']
        if self.type == 'dungeon':
            self.map = location['map']
            self.coord = location['coord']
            self.bossroad = location['bossroad']
        if self.type.startswith('dungeon') and len(self.type) > 7:
            self.dungeon = location['dungeon']
        if self.type == 'area' and 'dungeon' in location:
            self.dungeon = location['dungeon']
        self.link = location['link'].copy()
        self.visible = (
            location['visible'].copy() if 'visible' in location else {})
        if self.type not in ('area', 'interior'):
            self.checked = False
        else:
            self.checked = None


class Dungeon(dict):
    '''
    Dungeon object
    '''

    def keylocations(self) -> dict:
        '''
        Retrieve list oc locations possibly holding a small key.

        Returns:
            dict: {'location name': location object}
        '''

        ret = {}
        for loc in self:
            if self[loc].type  == 'dungeonchest':
                ret[loc] = self[loc]
        return ret



'''
Dark World map
'''

from . import maps

__all__ = 'start',


def start(worldtracker) -> maps.MapDisplay:
    '''
    Open dark world map.

    Args:
        worldtracker: world state tracker
    '''

    display = maps.MapDisplay('dark')
    worldtracker.add_map('dark', display)
    return display



'''
Item/dungeon tracker saving
'''

import json
import os.path

from ..config import CONFIG, CONFIGDIRECTORY
from ..version import __version__ as version

__all__ = (
    'load_items', 'load_dungeons', 'load_locations', 'load_entrances'
    'store_items', 'store_dungeons', 'store_locations','store_entrances')


def load_items() -> dict:
    '''
    Load item tracker data from save file.

    Returns:
        dict: save data
    '''

    try:
        ret = _load_save()['Items']
    except KeyError:
        ret = {}
    return ret


def load_dungeons() -> dict:
    '''
    Load dungeon tracker data from save file.

    Returns:
        dict: save data
    '''

    try:
        ret = _load_save()['Dungeons']
    except KeyError:
        ret = {}
    return ret


def load_locations() -> dict:
    '''
    Load location tracker data from save file.

    Returns:
        dict: save data
    '''

    try:
        ret = _load_save()['Locations']
    except KeyError:
        ret = {}
    return ret


def load_entrances() -> dict:
    '''
    Load entrance tracker data from save file.

    Returns:
        dict: save data
    '''

    try:
        ret = _load_save()['Entrances']
    except KeyError:
        ret = {}
    return ret


def _load_save() -> dict:
    '''
    Load save file.

    Return:
        dict: save data
    '''

    try:
        fid = open(os.path.join(CONFIGDIRECTORY, CONFIG['autosave']), 'r')
    except FileNotFoundError:
        return {}

    try:
        data = json.load(fid)
    except json.JSONDecodeError:
        return {}
    finally:
        fid.close()

    if data['version'] != version:
        return {}

    return data


def store_items(data: dict) -> None:
    '''
    Store item tracker state.

    Args:
        data: save data
    '''

    _store_save({'Items': data})


def store_dungeons(data: dict) -> None:
    '''
    Store dungeon tracker state.

    Args:
        data: save data
    '''

    _store_save({'Dungeons': data})


def store_locations(data: dict) -> None:
    '''
    Store location tracker state.

    Args:
        data: save data
    '''

    _store_save({'Locations': data})


def store_entrances(data: dict) -> None:
    '''
    Store location tracker state.

    Args:
        data: save data
    '''

    _store_save({'Entrances': data})


def _store_save(data: dict) -> None:
    '''
    Write save file.

    Args:
        data: save data
    '''

    savedata = _load_save()
    for dtype in data:
        savedata[dtype] = data[dtype]
    savedata['version'] = version
    with open(os.path.join(CONFIGDIRECTORY, CONFIG['autosave']), 'w') as fid:
        json.dump(savedata, fid)



'''
Dungeon management
'''

import configparser
import importlib
import os.path
import json

from ..config.images import image
layoutstorage = importlib.import_module(
    '..gui-common.dungeons', package=__package__)

from .dungeonobj import DungeonObj
from .lists import DUNGEONS
from . import storage

__all__ = 'DungeonTracker',


def get_layout() -> dict:
    '''
    Load (or create) dungeon layout.

    Returns:
        dict: dungeon layout in format {identifier: (column, row)}
    '''

    try:
        layout = layoutstorage.load_dungeons()
    except (layoutstorage.NoConfig, configparser.Error):
        layout = {}
        for dungeon in DUNGEONS:
            layout[dungeon] = DUNGEONS[dungeon]
        layoutstorage.new_dungeon(layout)
        layout = layoutstorage.load_dungeons()
    return layout


class DungeonTracker(dict):
    '''
    Dungeon info tracker.
    '''

    def __init__(self, worldtracker):
        '''
        Args:
            worldtracker: world state tracker
        '''

        layout = get_layout()

        super().__init__()
        self.worldtracker = worldtracker
        self.crystal_requirement = {'tower': -1, 'ganon': -1}
        delayed_link = []
        for rawdungeon in DUNGEONS:
            dungeon = DungeonObj(
                rawdungeon, DUNGEONS[rawdungeon][0], DUNGEONS[rawdungeon][1],
                DUNGEONS[rawdungeon][2], DUNGEONS[rawdungeon][3], worldtracker)
            try:
                dungeon.location = layout[dungeon]
            except KeyError:
                pass
            self[dungeon.identifier] = dungeon
        data = storage.load_dungeons()
        self.restore(data)

    def set_crystal(self, ctype: str, diff: int) -> int:
        '''
        Set crystal requirements.

        Args:
            ctype: 'tower' or 'crystal'
            diff: -1, +1 or None (for reset)
        Returns:
            int: number of required crystals, -1 means unknown
        '''

        if diff is None:
            self.crystal_requirement[ctype] = -1
        else:
            self.crystal_requirement[ctype] += diff
        if self.crystal_requirement[ctype] < -1:
            self.crystal_requirement[ctype] = 7
        elif self.crystal_requirement[ctype] > 7:
            self.crystal_requirement[ctype] = -1
        self.worldtracker.set_crystal_requirement(
            self.crystal_requirement['tower'],
            self.crystal_requirement['ganon'])
        self.save()
        return self.crystal_requirement[ctype]

    def reset(self) -> None:
        '''
        Reset all dungeons.
        '''

        for dungeon in self:
            self[dungeon].reset()

    def save(self, *args) -> dict:
        '''
        Save current dungeon setup.
        '''

        storage.store_dungeons(self.store())

    def store(self) -> dict:
        '''
        Return current dungeon info for storage.

        Returns:
            dict: dungeon info
        '''

        dungeoninfo = {}
        for dungeon in self:
            dungeoninfo[dungeon] = self[dungeon].save()
        dungeoninfo['crystal_requirements'] = self.crystal_requirement
        return dungeoninfo

    def restore(self, dungeoninfo) -> None:
        '''
        Restore current dungeon info from file.

        Args:
            dungeoninfo: information from file
        '''

        for dungeon in dungeoninfo:
            if dungeon in self:
                self[dungeon].restore(dungeoninfo[dungeon])
        try:
            self.crystal_requirement.update(dungeoninfo['crystal_requirements'])
        except KeyError:
            pass



'''
Location tracker
'''

from . import storage

__all__ = 'LocationTracker',


class LocationTracker(dict):
    '''
    Location tracker
    '''

    def __init__(self):
        super().__init__(storage.load_locations())

    def save(self) -> None:
        '''
        Save current location states.
        '''

        storage.store_locations(self)

    def reset(self) -> None:
        '''
        Reset tracker.
        '''

        for loc in self:
            if isinstance(self[loc], list):
                self[loc] = [True, True]
            else:
                self[loc] = True



'''
Config window
'''

import sys
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.messagebox as tkmbox
import typing

from ..config import CONFIG
from ..config.default import DEFAULT

from . import misc

SPACERWIDTH = -5
PADDING = 3

__all__ = 'start',


class ConfigWindow(tk.Toplevel):
    '''
    Config window

    Instance variables:
        tracker: world state tracker
        widgets: random collection of GUI widgets
        entries: current string config choices
        boolentries: current boolean config choices
        texts: texts display for config options
    '''

    def __init__(self, tracker):
        '''
        Args:
            tracker: world state tracker
        '''

        super().__init__()
        self.title('Settings')
        self.frame = ttk.Frame(self, padding=PADDING)
        self.frame.grid(column=0, row=0, sticky=misc.A)

        self.tracker = tracker

        self.widgets = []
        self.entries = {}
        self.boolentries = {}
        self.texts = {}

        col = ttk.Frame(self.frame)
        col.grid(column=0, row=0, sticky=tk.E+tk.W+tk.N)
        self.widgets.append(col)

        sec = ttk.LabelFrame(
            col, padding=PADDING,
            text='Program settings (requires restart)')
        sec.grid(column=0, row=0, sticky=tk.E+tk.W)
        self.widgets.append(sec)

        self._make_entry(0, sec, 'Item/Dungeon tracker size', 'icon_size')
        self._make_entry(1, sec, 'Maps size', 'map_size')

        sec = ttk.LabelFrame(
            col, padding=PADDING, text='File information')
        sec.grid(column=0, row=1, sticky=tk.E+tk.W)
        self.widgets.append(sec)

        self._make_display(0, sec, 'GUI module', 'gui')
        self._make_display(1, sec, 'Autosave', 'autosave')
        self._make_display(
            2, sec, 'Item/Dungeon tracker layout', 'button_layout')
        self._make_display(
            3, sec, 'Window layout', 'window_layout')

        col = ttk.Frame(self.frame)
        col.grid(column=1, row=0, sticky=tk.E+tk.W+tk.N)
        self.widgets.append(col)

        sec = ttk.LabelFrame(
            col, padding=PADDING, text='Game settings')
        sec.grid(column=0, row=0, sticky=tk.E+tk.W)
        self.widgets.append(sec)

        self._make_check(0, sec, 'Entrance randomiser', 'entrance_randomiser')
        self._make_choice(
            1, sec, 'World State', ('Standard', 'Open', 'Inverted', 'Retro'),
            'world_state')
        self._make_choice(
            2, sec, 'Glitches Required',
            ('None', 'Overworld Glitches', 'Major Glitches'), 'glitch_mode')
        self._make_choice(
            3, sec, 'Item Placement', ('Basic', 'Advanced'), 'item_placement')
        self._make_choice(
            4, sec, 'Dungeon Items',
            ('Standard', 'Maps/Compasses', 'Maps/Compasses/Small Keys',
             'Keysanity'),
            'dungeon_items')
        self._make_choice(
            5, sec, 'Goal',
            ('Defeat Ganon', 'Fast Ganon', 'All Dungeons',
             'Master Sword Pedestal', 'Triforce Hunt'),
            'goal')
        self._make_choice(
            6, sec, 'Swords',
            ('Randomised', 'Assured', 'Vanilla', 'Swordless'),
            'swords')

        buttonframe = ttk.Frame(self.frame)
        buttonframe.grid(column=0, columnspan=9, row=9, sticky=tk.E+tk.S)
        okbutton = ttk.Button(buttonframe, command=self.apply, text='Apply')
        cancelbutton = ttk.Button(
            buttonframe, command=self.withdraw, text='Cancel')
        col = 1 if sys.platform.startswith('win32') else 0
        okbutton.grid(column=col, row=0, sticky=tk.E)
        cancelbutton.grid(column=1 - col, row=0, sticky=tk.E)
        self.widgets.extend((buttonframe, okbutton, cancelbutton))

    def _make_entry(
            self, location: int, parent: ttk.LabelFrame, displaytext: str,
            configoption: str):
        '''
        Make text box.

        Args:
            location: row on GUI
            parent: parent widgets
            displaytext: text to show in config window
            configoption: corresponding config option
        '''

        name = ttk.Label(parent, text=displaytext)
        name.grid(column=0, row=location, sticky=tk.W)
        spacer = ttk.Label(parent, width=SPACERWIDTH)
        spacer.grid(column=1, row=location, sticky=tk.W)
        entryvar = tk.StringVar()
        entryvar.set(str(CONFIG[configoption]))
        entry = ttk.Entry(parent, textvariable=entryvar)
        entry.grid(column=2, row=location, sticky=tk.W)
        self.entries[configoption] = entryvar
        self.texts[configoption] = displaytext
        self.widgets.extend((name, spacer, entry))

    def _make_display(
            self, location: int, parent: ttk.LabelFrame, displaytext: str,
            configoption: str) -> None:
        '''
        Make info display.

        Args:
            location: row on GUI
            parent: parent widgets
            displaytext: text to show in config window
            configoption: corresponding config option
        '''

        name = ttk.Label(parent, text=displaytext)
        name.grid(column=0, row=location, sticky=tk.W)
        spacer = ttk.Label(parent, width=SPACERWIDTH)
        spacer.grid(column=1, row=location, sticky=tk.W)
        entryvar = tk.StringVar()
        entryvar.set(str(CONFIG[configoption]))
        entry = ttk.Entry(parent, textvariable=entryvar)
        entry.configure(state='readonly')
        entry.grid(column=2, row=location, sticky=tk.W)
        self.widgets.extend((name, spacer, entryvar, entry))

    def _make_check(
            self, location: int, parent: ttk.LabelFrame, displaytext: str,
            configoption: str) -> None:
        '''
        Make checkbox.

        Args:
            location: row on GUI
            parent: parent widgets
            displaytext: text to show in config window
            configoption: corresponding config option
        '''

        name = ttk.Label(parent, text=displaytext)
        name.grid(column=0, row=location, sticky=tk.W)
        spacer = ttk.Label(parent, width=SPACERWIDTH)
        spacer.grid(column=1, row=location, sticky=tk.W)
        entryvar = tk.IntVar()
        entryvar.set(int(CONFIG[configoption]))
        entry = ttk.Checkbutton(parent, variable=entryvar)
        entry.grid(column=2, row=location, sticky=tk.W)
        self.boolentries[configoption] = entryvar
        self.texts[configoption] = displaytext
        self.widgets.extend((name, spacer, entry))

    def _make_choice(
            self, location: int, parent: ttk.LabelFrame, displaytext: str,
            choices: typing.Sequence[str], configoption: str) -> None:
        '''
        Make menu selection.

        Args:
            location: row on GUI
            parent: parent widgets
            displaytext: text to show in config window
            choices: available selections
            configoption: corresponding config option
        '''

        name = ttk.Label(parent, text=displaytext)
        name.grid(column=0, row=location, sticky=tk.W)
        spacer = ttk.Label(parent, width=SPACERWIDTH)
        spacer.grid(column=1, row=location, sticky=tk.W)
        entryvar = tk.StringVar()
        entryvar.set(str(CONFIG[configoption]))
        entry = ttk.Combobox(parent, values=choices, textvariable=entryvar,
                             width=max(len(s) for s in choices))
        entry.configure(state='readonly')
        entry.grid(column=2, row=location, sticky=tk.W+tk.E)
        self.entries[configoption] = entryvar
        self.texts[configoption] = displaytext
        self.widgets.extend((name, spacer, entry))

    def withdraw(self) -> None:
        '''
        Reset all entries before closing the window.
        '''

        for entry in self.entries:
            self.entries[entry].set(str(CONFIG[entry]))
        for entry in self.boolentries:
            self.boolentries[entry].set(int(CONFIG[entry]))
        super().withdraw()

    def apply(self) -> None:
        '''
        Apply changed config.
        '''

        for entry in self.boolentries:
            CONFIG.update(
                entry, DEFAULT[entry][1](self.boolentries[entry].get()))
        for entry in self.entries:
            try:
                CONFIG.update(
                    entry, DEFAULT[entry][1](self.entries[entry].get()))
            except ValueError:
                tkmbox.showerror(
                    'Invalid',
                    "You have entered an incorrect value into '{0:s}'.".format(
                        self.texts[entry]),
                    parent=self)
                break
        else:
            self.tracker.reapply()
            self.dungeondisplay.redraw()
            super().withdraw()


def start(tracker) -> None:
    '''
    Open config window.

    Args:
        tracker: world state tracker
    '''

    return ConfigWindow(tracker)



'''
Skull Woods
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Kakariko Portal (DW)': {
        'type': 'area',
        'link': {
            'Kakariko Portal (LW)': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Kakariko Portal': [('item', 'powerglove')],
            'Skull Woods Surface': [('rabbitbarrier', None)],
            "Thieves' Town Surface": [('item', 'titansmitts')]}
    },
    'Skull Woods Surface': {
        'type': 'area',
        'link': {
            'Lost Woods': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Dark Bottle Ocarina': [],
            'Kakariko Portal (DW)': [],
            "Thieves' Town Surface": [],
            'Skull Woods First Entrance Entrance (E)': [],
            'Skull Woods West Drop Entrance (E)': [],
            'Skull Woods East Drop Entrance (E)': [],
            'Skull Woods Central Drop Entrance (E)': [('rabbitbarrier', None)],
            'Skull Woods Second Entrance Entrance (E)': [],
            'North Dark World': []}
    },
    'Skull Woods First Entrance Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (116, 98),
        'link': {
            'Skull Woods Surface': [],
            'Skull Woods First Entrance Entrance (I)': []}
    },
    'Skull Woods First Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Skull Woods First Entrance Entrance (E)': [],
            'Skull Woods First Entrance': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Skull Woods First Entrance': {
        'type': 'area', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods First Entrance Entrance (I)': [],
            'Skull Woods Map': [('rabbitbarrier', None)],
            'Skull Woods West Drop': [('smallkey', ('Skull Woods', 5))]}
    },
    'Skull Woods Map': {
        'type': 'dungeonchest', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods First Entrance': []}
    },
    'Skull Woods West Drop': {
        'type': 'area', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods First Entrance': [('smallkey', ('Skull Woods', 5))],
            'Skull Woods Front Key': [('rabbitbarrier', None)],
            'Skull Woods Compass': [('rabbitbarrier', None)],
            'Skull Woods East Drop': []}
    },
    'Skull Woods Front Key': {
        'type': 'dungeonchest', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods West Drop': []}
    },
    'Skull Woods Compass': {
        'type': 'dungeonchest', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods West Drop': []}
    },
    'Skull Woods East Drop': {
        'type': 'area', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods First Entrance': [('smallkey', ('Skull Woods', 5))],
            'Skull Woods Safety Key': [('rabbitbarrier', None)]}
    },
    'Skull Woods Safety Key': {
        'type': 'dungeonchest', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods East Drop': []}
    },
    'Skull Woods West Drop Entrance (E)': {
        'type': 'entrance_drop', 'map': 'dark', 'coord': (98, 117),
        'link': {
            'Skull Woods West Drop Entrance (I)': []}
    },
    'Skull Woods West Drop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Skull Woods West Drop': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Skull Woods East Drop Entrance (E)': {
        'type': 'entrance_drop', 'map': 'dark', 'coord': (124, 113),
        'link': {
            'Skull Woods East Drop Entrance (I)': []}
    },
    'Skull Woods East Drop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Skull Woods East Drop': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Skull Woods Central Drop Entrance (E)': {
        'type': 'entrance_drop', 'map': 'dark', 'coord': (120, 83),
        'link': {
            'Skull Woods Central Drop Entrance (I)': []}
    },
    'Skull Woods Central Drop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Skull Woods Central Drop': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Skull Woods Central Drop': {
        'type': 'area', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods First Entrance': [],
            'Skull Woods Treasure': [('rabbitbarrier', None)]}
    },
    'Skull Woods Treasure': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Central Drop': []}
    },
    'Skull Woods Second Entrance Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (90, 96),
        'link': {
            'Skull Woods Surface': [],
            'Skull Woods Second Entrance Entrance (I)': []}
    },
    'Skull Woods Second Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Skull Woods Second Entrance Entrance (E)': [],
            'Skull Woods Connector': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Skull Woods Connector': {
        'type': 'area', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Second Entrance Entrance (I)': [],
            'Skull Woods Big Key': [('rabbitbarrier', None)],
            'Skull Woods Hidden Key': [('rabbitbarrier', None)],
            'Skull Woods Third Entrance Entrance (I)': []}
    },
    'Skull Woods Big Key': {
        'type': 'dungeonchest', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Connector': []}
    },
    'Skull Woods Hidden Key': {
        'type': 'dungeonkey', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Connector': []}
    },
    'Skull Woods Third Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Skull Woods Connector': [],
            'Skull Woods Third Entrance Entrance (E)': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Skull Woods Third Entrance Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (33, 86),
        'link': {
            'Skull Woods Third Entrance Entrance (I)': [],
            'Hidden Skull Woods': []}
    },
    'Hidden Skull Woods': {
        'type': 'area',
        'link': {
            'Lost Woods': [('and', [
                ('nosettings', 'inverted'), ('item', 'mirror')])],
            'Skull Woods Third Entrance Entrance (E)': [],
            'Skull Woods North Drop Entrance (E)': [],
            'Skull Woods Fourth Entrance Entrance (E)': [('item', 'firerod')]}
    },
    'Skull Woods North Drop Entrance (E)': {
        'type': 'entrance_drop', 'map': 'dark', 'coord': (74, 60),
        'link': {
            'Skull Woods North Drop Entrance (I)': []}
    },
    'Skull Woods North Drop Entrance (I)': {
        'type': 'interior',
        'link': {
            'Skull Woods Connector': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Skull Woods Fourth Entrance Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (20, 43),
        'link': {
            'Hidden Skull Woods': [],
            'Skull Woods Fourth Entrance Entrance (I)': []}
    },
    'Skull Woods Fourth Entrance Entrance (I)': {
        'type': 'interior',
        'link': {
            'Skull Woods Fourth Entrance Entrance (E)': [],
            'Skull Woods Gauntlet Entrance': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Skull Woods Gauntlet Entrance': {
        'type': 'area', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Fourth Entrance Entrance (I)': [],
            'Skull Woods Gauntlet Key': [('rabbitbarrier', None)],
            'Skull Woods Gauntlet 1': [('smallkey', ('Skull Woods', 5))]}
    },
    'Skull Woods Gauntlet Key': {
        'type': 'dungeonchest', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Gauntlet Entrance': []}
    },
    'Skull Woods Gauntlet 1': {
        'type': 'area', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Gauntlet Entrance': [],
            'Skull Woods Gauntlet 2': [('and', [
                ('item', 'firerod'),
                ('item', 'sword')])]}
    },
    'Skull Woods Gauntlet 2': {
        'type': 'area', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Gauntlet 2': [],
            'Skull Woods Gibdo Key': [],
            'Skull Woods Boss': [('smallkey', ('Skull Woods', 5))]}
    },
    'Skull Woods Gibdo Key': {
        'type': 'dungeonkey', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Gauntlet 2': []}
    },
    'Skull Woods Boss': {
        'type': 'dungeonboss', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Boss Item': [('and', [
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('item', 'mastersword'),
                    ('and', [
                        ('or', [
                            ('item', 'bottle'), ('item', 'halfmagic')]),
                        ('item', 'firerod')])]),
                ('or', [
                    ('and', [
                        ('or', [
                            ('item', 'bottle'), ('item', 'halfmagic')]),
                        ('or', [
                            ('item', 'firerod'), ('item', 'somaria'),
                            ('item', 'byrna')])]),
                    ('item', 'sword'),
                    ('item', 'hammer')])])]}
    },
    'Skull Woods Boss Item': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Reward': []}
    },
    'Skull Woods Reward': {
        'type': 'dungeonreward', 'dungeon': 'Skull Woods',
        'link': {
            'Skull Woods Fourth Entrance Entrance (I)': []}
    },
}



'''
Dungeons
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Castle Dungeons': {
        'type': 'drop', 'map': 'light', 'coord': (332, 290),
        'link': {
            "Zelda's Treasure": []}
    },
    'East Palace': {
        'type': 'dungeon', 'map': 'light', 'coord': (636, 251),
        'bossroad': 'East Palace Boss',
        'link': {}
    },
    'Desert Palace': {
        'type': 'dungeon', 'map': 'light', 'coord': (49, 528),
        'bossroad': 'Desert Palace Trap Room 1',
        'link': {}
    },
    'Mountain Tower': {
        'type': 'dungeon', 'map': 'light', 'coord': (372, 14),
        'bossroad': 'Mountain Tower Interior',
        'link': {}
    },
    'Castle Tower': {
        'type': 'dungeon', 'map': 'light', 'coord': (332, 244),
        'bossroad': 'Castle Tower Boss',
        'link': {}
    },
    'Dark Palace': {
        'type': 'dungeon', 'map': 'dark', 'coord': (629, 241),
        'bossroad': 'Dark Palace Gauntlet',
        'link': {}
    },
    'Swamp Palace': {
        'type': 'dungeon', 'map': 'dark', 'coord': (304, 600),
        'bossroad': 'Swamp Palace Gauntlet',
        'link': {}
    },
    'Skull Woods': {
        'type': 'dungeon', 'map': 'dark', 'coord': (20, 22),
        'bossroad': 'Skull Woods Gauntlet 1',
        'link': {}
    },
    "Thieves' Town": {
        'type': 'dungeon', 'map': 'dark', 'coord': (77, 301),
        'bossroad': "Thieves' Town Blind's Cell",
        'link': {}
    },
    'Ice Palace': {
        'type': 'dungeon', 'map': 'dark', 'coord': (523, 552),
        'bossroad': 'Ice Palace Boss Puzzle',
        'link': {}
    },
    'Misery Mire': {
        'type': 'dungeon', 'map': 'dark', 'coord': (43, 532),
        'bossroad': 'Misery Mire Gauntlet',
        'link': {}
    },
    'Turtle Rock': {
        'type': 'dungeon', 'map': 'dark', 'coord': (618, 33),
        'bossroad': 'Turtle Rock Boss Door',
        'link': {}
    },
    "Ganon's Tower": {
        'type': 'dungeon', 'map': 'dark', 'coord': (367, 14),
        'bossroad': "Ganon's Tower Lobby",
        'link': {}
    },
}



'''
Ganon's Tower
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    "Ganon's Tower Entrance (I)": {
        'type': 'interior',
        'link': {
            'Castle Tower Entrance (E)': [('settings', 'inverted')],
            "Ganon's Tower Entrance (E)": [('nosettings', 'inverted')],
            "Ganon's Tower Lobby": [('and', [
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('and', [
                        ('or', [
                            ('settings', 'swordless'),
                            ('item', 'mastersword')]),
                        ('or', [
                            ('item', 'bottle'), ('item', 'bluemail')])])]),
                ('rabbitbarrier', None)])]}
    },
    "Ganon's Tower Lobby": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Entrance (I)": [],
            "Ganon's Tower Torch Key Room": [],
            "Ganon's Tower Trap Room": [],
            "Ganon's Tower Ascent 1": [('bigkey', "Ganon's Tower")]}
    },

    "Ganon's Tower Torch Key Room": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Lobby": [],
            "Ganon's Tower Torch Key": [('item', 'pegasus')],
            "Ganon's Tower Moving Bumper Key": []}
    },
    "Ganon's Tower Torch Key": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Torch Key Room": []}
    },
    "Ganon's Tower Moving Bumper Key": {
        'type': 'dungeonkey', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Torch Key Room": [],
            "Ganon's Tower Pit Room": [('item', 'hammer')]}
    },
    "Ganon's Tower Pit Room": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Moving Bumper Key": [('item', 'hookshot')],
            "Ganon's Tower Stalfos Room": [('item', 'hookshot')],
            "Ganon's Tower Map": [('item', 'hookshot'), ('item', 'pegasus')]}
    },
    "Ganon's Tower Stalfos Room": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Pit Room": [
                ('item', 'hookshot'), ('item', 'pegasus')],
            "Ganon's Tower Stalfos Room Chest 1": [],
            "Ganon's Tower Stalfos Room Chest 2": [],
            "Ganon's Tower Stalfos Room Chest 3": [],
            "Ganon's Tower Stalfos Room Chest 4": []}
    },
    "Ganon's Tower Stalfos Room Chest 1": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Stalfos Room": []}
    },
    "Ganon's Tower Stalfos Room Chest 2": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Stalfos Room": []}
    },
    "Ganon's Tower Stalfos Room Chest 3": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Stalfos Room": []}
    },
    "Ganon's Tower Stalfos Room Chest 4": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Stalfos Room": []}
    },
    "Ganon's Tower Map": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Pit Room": [('item', 'hookshot')],
            "Ganon's Tower Switch Key": []}
    },
    "Ganon's Tower Switch Key": {
        'type': 'dungeonkey', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Map": [],
            "Ganon's Tower Winder Room": [('item', 'hookshot')]}
    },
    "Ganon's Tower Winder Room": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Secret Treasure": [('item', 'bombs')],
            "Ganon's Tower Convergence": []}
    },
    "Ganon's Tower Secret Treasure": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Secret Chest 1": [],
            "Ganon's Tower Secret Chest 2": [],
            "Ganon's Tower Secret Chest 3": [],
            "Ganon's Tower Secret Chest 4": [],
            "Ganon's Tower Convergence": []}
    },
    "Ganon's Tower Secret Chest 1": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Secret Treasure": []}
    },
    "Ganon's Tower Secret Chest 2": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Secret Treasure": []}
    },
    "Ganon's Tower Secret Chest 3": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Secret Treasure": []}
    },
    "Ganon's Tower Secret Chest 4": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Secret Treasure": []}
    },

    "Ganon's Tower Trap Room": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Lobby": [],
            "Ganon's Tower Trap Chest 1": [],
            "Ganon's Tower Trap Chest 2": [],
            "Ganon's Tower Tile Room": [('item', 'somaria')]}
    },
    "Ganon's Tower Trap Chest 1": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Trap Room": []}
    },
    "Ganon's Tower Trap Chest 2": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Trap Room": []}
    },
    "Ganon's Tower Tile Room": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Trap Room": [],
            "Ganon's Tower Torch Race": []}
    },
    "Ganon's Tower Torch Race": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Tile Room": [],
            "Ganon's Tower Compass Room": [('item', 'firerod')]}
    },
    "Ganon's Tower Compass Room": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Torch Race": [],
            "Ganon's Tower Compass Chest 1": [],
            "Ganon's Tower Compass Chest 2": [],
            "Ganon's Tower Compass Chest 3": [],
            "Ganon's Tower Compass Chest 4": [],
            "Ganon's Tower Obstacle Course Key": []}
    },
    "Ganon's Tower Compass Chest 1": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Compass Room": []}
    },
    "Ganon's Tower Compass Chest 2": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Compass Room": []}
    },
    "Ganon's Tower Compass Chest 3": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Compass Room": []}
    },
    "Ganon's Tower Compass Chest 4": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Compass Room": []}
    },
    "Ganon's Tower Obstacle Course Key": {
        'type': 'dungeonkey', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Convergence": []}
    },

    "Ganon's Tower Convergence": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Anti-Fairy Room": [],
            "Ganon's Tower Treasure": []}
    },
    "Ganon's Tower Anti-Fairy Room": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Convergence": [],
            "Ganon's Tower Armos On Ice": [('item', 'bombs')]}
    },
    "Ganon's Tower Armos On Ice": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Big Key Room": [],
            "Ganon's Tower Treasure": []}
    },
    "Ganon's Tower Big Key Room": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Armos On Ice": [],
            "Ganon's Tower Big Key Chest 1": [],
            "Ganon's Tower Big Key Chest 2": [],
            "Ganon's Tower Big Key Chest 3": []}
    },
    "Ganon's Tower Big Key Chest 1": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Big Key Room": []}
    },
    "Ganon's Tower Big Key Chest 2": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Big Key Room": []}
    },
    "Ganon's Tower Big Key Chest 3": {
        'type': 'dungeonchest', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Big Key Room": []}
    },
    "Ganon's Tower Treasure": {
        'type': 'dungeonchest_nokey', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Convergence": [],
            "Ganon's Tower Torch Key Room": []}
    },

    "Ganon's Tower Ascent 1": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Lobby": [],
            "Ganon's Tower Ascent 2": [('item', 'bow')]}
    },
    "Ganon's Tower Ascent 2": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Ascent 1": [('item', 'bow')],
            "Ganon's Tower Ascent 3": [
                ('item', 'lantern'), ('item', 'firerod')]}
    },
    "Ganon's Tower Ascent 3": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Ascent 2": [],
            "Ganon's Tower Helmasaur Key": [],
            "Ganon's Tower Helmasaur Chest 1": [],
            "Ganon's Tower Helmasaur Chest 2": [],
            "Ganon's Tower Ascent 4": []}
    },
    "Ganon's Tower Helmasaur Key": {
        'type': 'dungeonkey', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Ascent 3": []}
    },
    "Ganon's Tower Helmasaur Chest 1": {
        'type': 'dungeonchest_nokey', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Ascent 3": []}
    },
    "Ganon's Tower Helmasaur Chest 2": {
        'type': 'dungeonchest_nokey', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Ascent 3": []}
    },
    "Ganon's Tower Ascent 4": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Ascent 3": [],
            "Ganon's Tower Rabbit Beam Chest": [],
            "Ganon's Tower Ascent 5": []}
    },
    "Ganon's Tower Rabbit Beam Chest": {
        'type': 'dungeonchest_nokey', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Ascent 4": []}
    },
    "Ganon's Tower Ascent 5": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Ascent 4": [],
            "Ganon's Tower Ascent 6": [('item', 'hookshot')]}
    },
    "Ganon's Tower Ascent 6": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Last Chest": [],
            "Ganon's Tower Boss": []}
    },
    "Ganon's Tower Last Chest": {
        'type': 'dungeonchest_nokey', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Ascent 6": []}
    },
    "Ganon's Tower Boss": {
        'type': 'dungeonboss', "dungeon": "Ganon's Tower",
        'link': {
            "Ganon's Tower Boss Item": [
                ('item', 'sword'),
                ('and', [
                    ('settings', 'swordless'), ('item', 'hammer')]),
                ('item', 'bugnet')]}
    },
    "Ganon's Tower Boss Item": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            "Ganon's Tower Reward": []}
    },
    "Ganon's Tower Reward": {
        'type': 'area', 'dungeon': "Ganon's Tower",
        'link': {
            'Pyramid': [('nosettings', 'inverted')],
            'Castle Walls': [('settings', 'inverted')]}
    },
}



'''
General dungeon info
'''

__all__ = 'DUNGEONS',


DUNGEONS = {
    'East Palace': (
        (0, 0), 0, 'eastpalace',
        ('map', 'compass', 'bigkey', 'reward')
    ),
    'Desert Palace': (
        (1, 0), 1, 'desertpalace',
        ('map', 'compass', 'bigkey', 'reward')
    ),
    'Mountain Tower': (
        (2, 0), 1, 'mountaintower',
        ('map', 'compass', 'bigkey', 'reward')
    ),
    'Castle Tower': (
        (3, 0), 2, 'castletower',
        ()
    ),
    'Dark Palace': (
        (0, 1), 6, 'darkpalace',
        ('map', 'compass', 'bigkey', 'reward')
    ),
    'Swamp Palace': (
        (1, 1), 1, 'swamppalace',
        ('map', 'compass', 'bigkey', 'reward')
    ),
    'Skull Woods': (
        (2, 1), 3, 'skullwoods',
        ('map', 'compass', 'bigkey', 'reward')
    ),
    "Thieves' Town": (
        (3, 1), 1, 'thievestown',
        ('map', 'compass', 'bigkey', 'reward')
    ),
    'Ice Palace': (
        (0, 2), 2, 'icepalace',
        ('map', 'compass', 'bigkey', 'reward')
    ),
    'Misery Mire': (
        (1, 2), 3, 'miserymire',
        ('map', 'compass', 'bigkey', 'reward', 'medallion')
    ),
    'Turtle Rock': (
        (2, 2), 4, 'turtlerock',
        ('map', 'compass', 'bigkey', 'reward', 'medallion')
    ),
    "Ganon's Tower": (
        (3, 2), 4, 'ganonstower',
        ('map', 'compass', 'bigkey')
    )
}



'''
Swamp Palace
'''

__all__ = 'LOCATIONS',


LOCATIONS = {
    'Swamp Palace Entrance (E)': {
        'type': 'entrance_dungeon', 'map': 'dark', 'coord': (304, 621),
        'link': {
            'South Dark World': [],
            'Swamp Palace Entrance (I)': []}
    },
    'Swamp Palace Entrance (I)': {
        'type': 'interior',
        'link': {
            'Swamp Palace Entrance (E)': [],
            'Swamp Palace Before Canal': [
                ('settings', 'placement_advanced'),
                ('and', [
                    ('or', [
                        ('settings', 'swordless'), ('item', 'sword')]),
                    ('item', 'bottle')])]}
    },
    'Swamp Palace Before Canal': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Entrance (I)': [],
            'Swamp Palace Canal': [('and', [
                ('access', 'Water Drain'),
                ('or', [('settings', 'entrance'), ('item', 'mirror')])])]}
    },
    'Swamp Palace Canal': {
        'type': 'area', 'dungeon': 'Swamp Palace After Canal',
        'link': {
            'Swamp Palace Before Canal': [],
            'Swamp Palace After Canal': [('item', 'flippers')]}
    },
    'Swamp Palace After Canal': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Canal': [('item', 'flippers')],
            'Swamp Palace Key': [],
            'Swamp Palace Water Hall': [('smallkey', ('Swamp Palace', 1))]}
    },
    'Swamp Palace Key': {
        'type': 'dungeonchest', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace After Canal': []}
    },
    'Swamp Palace Water Hall': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace After Canal': [],
            'Swamp Palace Map': [('item', 'bombs')],
            'Swamp Palace First Hidden Key': [],
            'Swamp Palace East Basin': [('smallkey', ('Swamp Palace', 1))]}
    },
    'Swamp Palace Map': {
        'type': 'dungeonchest', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Water Hall': []}
    },
    'Swamp Palace First Hidden Key': {
        'type': 'dungeonkey', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Water Hall': []}
    },
    'Swamp Palace East Basin': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Water Hall': [],
            'Swamp Palace Second Hidden Key': [],
            'Swamp Palace East Switch Room': [
                ('smallkey', ('Swamp Palace', 1))]}
    },
    'Swamp Palace Second Hidden Key': {
        'type': 'dungeonkey', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace East Basin': []}
    },
    'Swamp Palace East Switch Room': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace East Basin': [],
            'Swamp Palace East Switch': [('item', 'hammer')]}
    },
    'Swamp Palace East Switch': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace East Switch Room': [],
            'Swamp Palace Water Hall': [('item', 'flippers')],
            'Swamp Palace Main Hall': [('item', 'flippers')]}
    },
    'Swamp Palace Main Hall': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace East Switch': [],
            'Swamp Palace Compass': [],
            'Swamp Palace Third Hidden Key': [],
            'Swamp Palace West Switch': [('smallkey', ('Swamp Palace', 1))],
            'Swamp Palace Treasure': [('bigkey', 'Swamp Palace')],
            'Swamp Palace Fourth Hidden Key': [('item', 'hookshot')],
            'Swamp Palace Gauntlet Entrance': [('item', 'hookshot')]}
    },
    'Swamp Palace Compass': {
        'type': 'dungeonchest', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Main Hall': []}
    },
    'Swamp Palace Third Hidden Key': {
        'type': 'dungeonkey', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Main Hall': []}
    },
    'Swamp Palace West Switch': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Main Hall': [],
            'Swamp Palace West Basin': [('item', 'flippers')]}
    },
    'Swamp Palace West Basin': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Main Hall': [('item', 'flippers')],
            'Swamp Palace Fake Big Key': [],
            'Swamp Palace Big Key': []}
    },
    'Swamp Palace Fake Big Key': {
        'type': 'dungeonchest', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace West Basin': []}
    },
    'Swamp Palace Big Key': {
        'type': 'dungeonchest', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace West Basin': []}
    },
    'Swamp Palace Treasure': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Main Hall': []}
    },
    'Swamp Palace Fourth Hidden Key': {
        'type': 'dungeonkey', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Main Hall': []}
    },
    'Swamp Palace Gauntlet Entrance': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Main Hall': [],
            'Swamp Palace Gauntlet': [('smallkey', ('Swamp Palace', 1))]}
    },
    'Swamp Palace Gauntlet': {
        'type': 'area', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Gauntlet Entrance': [],
            'Swamp Palace Left Treasury Chest': [],
            'Swamp Palace Right Treasury Chest': [],
            'Swamp Palace Hidden Chest': [],
            'Swamp Palace Boss': [('item', 'flippers')]}
    },
    'Swamp Palace Left Treasury Chest': {
        'type': 'dungeonchest', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Gauntlet': []}
    },
    'Swamp Palace Right Treasury Chest': {
        'type': 'dungeonchest', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Gauntlet': []}
    },
    'Swamp Palace Hidden Chest': {
        'type': 'dungeonchest', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Gauntlet': []}
    },
    'Swamp Palace Boss': {
        'type': 'dungeonboss', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Boss Item': [('and', [
                ('item', 'hookshot'),
                ('or', [
                    ('settings', 'placement_advanced'),
                    ('settings', 'swordless'),
                    ('item', 'mastersword')]),
                ('or', [
                    ('item', 'hammer'),
                    ('item', 'sword'),
                    ('and', [
                        ('or', [
                            ('item', 'halfmagic'), ('item', 'bottle'),
                            ('item', 'bow')]),
                        ('or', [
                            ('item', 'firerod'), ('item', 'icerod')])])])])]}
    },
    'Swamp Palace Boss Item': {
        'type': 'dungeonchest_nokey', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Reward': []}
    },
    'Swamp Palace Reward': {
        'type': 'dungeonreward', 'dungeon': 'Swamp Palace',
        'link': {
            'Swamp Palace Entrance (I)': []}
    },
}
